import abc

from audiomate.corpus import base


class FilterCriterion(metaclass=abc.ABCMeta):
    """
    A filter criterion decides wheter a given utterance contained in a given corpus matches the
    filter.
    """

    def match(self, utterance, corpus):
        """
        Check if the utterance matches the filter.

        Args:
            utterance (Utterance): The utterance to match.
            corpus (CorpusView): The corpus that contains the utterance.

        Returns:
            bool: True if the filter matches the utterance, False otherwise.
        """
        pass

    @abc.abstractmethod
    def serialize(self):
        """
        Serialize this filter criterion to write to a file.
        The output needs to be a single line without line breaks.

        Returns:
            str: A string representing this filter criterion.
        """
        pass

    @classmethod
    @abc.abstractmethod
    def parse(cls, representation):
        """
        Create a filter criterion based on a string representation (created with ``serialize``).

        Args:
            representation (str): The string representation.

        Returns:
            FilterCriterion: The filter criterion from that representation.
        """
        pass

    @classmethod
    @abc.abstractmethod
    def name(cls):
        """
        Returns a name identifying this type of filter criterion.
        """
        return 'unknown'


class MatchingUtteranceIdxFilter(FilterCriterion):
    """
    A filter criterion that matches utterances based on utterance-ids.

    Args:
        utterance_idxs (:class:`set`): A list of utterance-ids. Only utterances in the list will pass the
                               filter
        inverse (bool): If True only utterance not in the list pass the filter.
    """

    def __init__(self, utterance_idxs=set(), inverse=False):
        self.utterance_idxs = set(utterance_idxs)
        self.inverse = inverse

    def match(self, utterance, corpus):
        return (not self.inverse and utterance.idx in self.utterance_idxs) \
            or (self.inverse and utterance.idx not in self.utterance_idxs)

    def serialize(self):
        inverse_indication = 'exclude' if self.inverse else 'include'
        id_string = ','.join(sorted(self.utterance_idxs))
        return '{},{}'.format(inverse_indication, id_string)

    @classmethod
    def parse(cls, representation):
        items = representation.strip().split(',')
        inverse_indication = items.pop(0)
        inverse = inverse_indication == 'exclude'

        return cls(utterance_idxs=set(items), inverse=inverse)

    @classmethod
    def name(cls):
        return 'matching_utterance_ids'


class MatchingLabelFilter(FilterCriterion):
    """
    A filter criterion that only accepts utterances which only have the given labels.

    Args:
        labels (:class:`set`): A set of labels which are accepted.
        label_list_ids (:class:`set`): Only check label-lists with these ids. If empty checks all label-lists.
    """

    def __init__(self, labels=set(), label_list_ids=set()):
        self.labels = labels
        self.label_list_ids = label_list_ids

    def match(self, utterance, corpus):
        for label_list_idx, label_list in utterance.label_lists.items():
            if len(self.label_list_ids) == 0 or label_list_idx in self.label_list_ids:
                for label in label_list:
                    if label.value not in self.labels:
                        return False

        return True

    def serialize(self):
        ll_ids = ','.join(sorted(self.label_list_ids))
        labels = ','.join(sorted(self.labels))
        return '{}|||{}'.format(ll_ids, labels)

    @classmethod
    def parse(cls, representation):
        parts = representation.strip().split('|||')

        if len(parts) > 1:
            ll_ids = parts[0].strip().split(',')
            labels = parts[1].strip().split(',')
        else:
            ll_ids = set()
            labels = parts[0].strip().split(',')

        return cls(set(labels), set(ll_ids))

    @classmethod
    def name(cls):
        return 'matching_labels'


__filter_criteria = {}
for cls in FilterCriterion.__subclasses__():
    __filter_criteria[cls.name()] = cls


class UnknownFilterCriteriaException(Exception):
    pass


def available_filter_criteria():
    """
    Get a mapping of all available filter criteria.

    Returns:
        dict: A dictionary with filter-criterion classes with the name of these criteria as key.

    Example::

        >>> available_filter_criteria()
        {
            "matching_utterance_ids" : audiomate.corpus.subview.MatchingUtteranceIdxFilter
        }
    """
    return __filter_criteria


class Subview(base.CorpusView):
    """
    A subview is a readonly layer representing some subset of a corpus.
    The assets the subview contains are defined by filter criteria.
    Only if an utterance passes all filter criteria it is contained in the subview.

    Args:
        corpus (CorpusView): The corpus this subview is based on.
        filter_criteria (list, FilterCriterion): List of :py:class:`FilterCriterion`

    Example::

        >>> filter = subview.MatchingUtteranceIdxFilter(utterance_idxs=(['utt-1', 'utt-3']))
        >>> corpus = audiomate.corpus.load('path/to/corpus')
        >>> corpus.num_utterances
        14
        >>> subset = subview.Subview(self.corpus, filter_criteria=[filter])
        >>> subset.num_utterances
        2
    """

    def __init__(self, corpus, filter_criteria=[]):
        self.corpus = corpus

        if isinstance(filter_criteria, list):
            self.filter_criteria = filter_criteria
        else:
            self.filter_criteria = [filter_criteria]

        self._cached_utterances = None
        self._all_utterance_ids = None

    @property
    def name(self):
        return 'subview of {}'.format(self.corpus.name)

    @property
    def tracks(self):
        return {utterance.track.idx: utterance.track for utterance in self.utterances.values()}

    @property
    def utterances(self):
        # Check if filtered utterances are cached
        # and return them if utterances in parent corpus haven't changed
        if self._cached_utterances is not None:
            if not set(self.corpus.utterances.keys()).isdisjoint(self._all_utterance_ids):
                return self._cached_utterances

        filtered_utterances = {}
        idx_utterances = self.corpus.utterances.items()

        for utt_idx, utterance in idx_utterances:
            matches = True

            for criterion in self.filter_criteria:
                if not criterion.match(utterance, self.corpus):
                    matches = False

            if matches:
                filtered_utterances[utt_idx] = utterance

        # Cache filtered utterances
        self._cached_utterances = filtered_utterances
        self._all_utterance_ids = {u[0] for u in idx_utterances}

        return filtered_utterances

    @property
    def issuers(self):
        issuers = {}

        for utterance in self.utterances.values():
            if utterance.issuer is not None:
                issuers[utterance.issuer.idx] = utterance.issuer

        return issuers

    @property
    def feature_containers(self):
        return self.corpus.feature_containers

    def serialize(self):
        """
        Return a string representing the subview with all of its filter criteria.

        Returns:
            str: String with subview definition.
        """
        lines = []

        for criterion in self.filter_criteria:
            lines.append(criterion.name())
            lines.append(criterion.serialize())

        return '\n'.join(lines)

    @classmethod
    def parse(cls, representation, corpus=None):
        """
        Creates a subview from a string representation (created with ``self.serialize``).

        Args:
            representation (str): The representation.

        Returns:
            Subview: The created subview.
        """

        criteria_definitions = representation.split('\n')
        criteria = []

        for i in range(0, len(criteria_definitions), 2):
            filter_name = criteria_definitions[i]
            filter_repr = criteria_definitions[i + 1]

            if filter_name not in available_filter_criteria():
                raise UnknownFilterCriteriaException('Unknown filter-criterion {}'.format(filter_name))

            criterion = available_filter_criteria()[filter_name].parse(filter_repr)
            criteria.append(criterion)

        return cls(corpus, criteria)



import librosa

from . import base


class PowerToDb(base.Computation):
    """
    Convert a power spectrogram (amplitude squared) to decibel (dB) units.

    See http://librosa.github.io/librosa/generated/librosa.core.power_to_db.html

    Note:

        The output can differ depending on offline or online processing, since it depends on statistics over all values.
        And in online mode it only considers values from a single chunk,
        while in offline mode all values of the whole sequence are considered.
    """

    def __init__(self, ref=1.0, amin=1e-10, top_db=80.0, parent=None, name=None):
        super(PowerToDb, self).__init__(parent=parent, name=name)

        self.ref = ref
        self.amin = amin
        self.top_db = top_db

    def compute(self, chunk, sampling_rate, corpus=None, utterance=None):
        return librosa.power_to_db(chunk.data.T, ref=self.ref, amin=self.amin, top_db=self.top_db).T



"""
This module contains classes for creating frame processing pipelines.

A pipeline consists of one of two types of steps. A computation step takes data from a previous step or the input and
processes it. A reduction step is used to merge outputs of multiple previous steps.
It takes outputs of all incoming steps and outputs a single data block.

The steps are managed as a directed graph,
which is built by passing the parent steps to the ``__init__`` method of a step.
Every step that is created has his own graph, but inherits all nodes and edges of the graphs of his parent steps.

Every pipeline represents a processor and implements the ``process_frames`` method.
"""

from .base import Chunk  # noqa: F401
from .base import Step  # noqa: F401
from .base import Computation  # noqa: F401
from .base import Reduction  # noqa: F401

from .normalization import MeanVarianceNorm  # noqa: F401

from .spectral import MelSpectrogram  # noqa: F401
from .spectral import MFCC  # noqa: F401

from .magnitude_scaling import PowerToDb  # noqa: F401

from .varia import Delta  # noqa: F401
from .varia import AddContext  # noqa: F401

from .reduction import Stack  # noqa: F401

from .pool import AvgPool  # noqa: F401
from .pool import VarPool  # noqa: F401

from .onset import OnsetStrength  # noqa: F401
from .rhythm import Tempogram  # noqa: F401



import numpy as np
import librosa
import pytest

from audiomate import tracks
from audiomate import annotations
from audiomate import issuers

from tests import resources


class TestUtterance:

    def setup_method(self):
        self.ll_1 = annotations.LabelList(idx='alpha', labels=[
            annotations.Label('a', 3.2, 4.5),
            annotations.Label('b', 5.1, 8.9),
            annotations.Label('c', 7.2, 10.5),
            annotations.Label('d', 10.5, 14),
            annotations.Label('d', 15, 18)
        ])

        self.ll_2 = annotations.LabelList(idx='bravo', labels=[
            annotations.Label('a', 1.0, 4.2),
            annotations.Label('e', 4.2, 7.9),
            annotations.Label('c', 7.2, 10.5),
            annotations.Label('f', 10.5, 14),
            annotations.Label('d', 15, 17.3)
        ])

        self.ll_duplicate_idx = annotations.LabelList(idx='charlie', labels=[
            annotations.Label('t', 1.0, 4.2),
            annotations.Label('h', 4.2, 7.9)
        ])

        self.ll_3 = annotations.LabelList(idx='charlie', labels=[
            annotations.Label('a', 1.0, 4.2),
            annotations.Label('g', 4.2, 7.9)
        ])

        self.track = tracks.FileTrack('wav', resources.sample_wav_file('wav_1.wav'))
        self.issuer = issuers.Issuer('toni')
        self.utt = tracks.Utterance(
            'test',
            self.track,
            issuer=self.issuer,
            start=1.25,
            end=1.30,
            label_lists=[
                self.ll_1,
                self.ll_2,
                self.ll_duplicate_idx,
                self.ll_3
            ]
        )

    def test_end_abs(self):
        assert self.utt.end_abs == 1.30

    def test_end_abs_end_of_track(self):
        utt = tracks.Utterance('utt', self.track, start=0.3)
        assert utt.end_abs == pytest.approx(2.5951875)

    def test_duration(self):
        assert self.utt.duration == pytest.approx(0.05)

    def test_duration_end_of_track(self):
        utt = tracks.Utterance('utt', self.track, start=0.3)
        assert utt.duration == pytest.approx(2.2951875)

    def test_duration_longer_than_track(self):
        utt = tracks.Utterance('utt', self.track, start=0.3, end=255.0)
        assert utt.duration == pytest.approx(254.7)

    def test_issuer_relation_on_creation(self):
        assert self.utt.issuer == self.issuer
        assert self.utt in self.issuer.utterances

    def test_set_label_list(self):
        ll_4 = annotations.LabelList(idx='delta', labels=[
            annotations.Label('y', 0.0, 3.3),
            annotations.Label('t', 3.8, 7.9)
        ])

        self.utt.set_label_list(ll_4)

        assert len(self.utt.label_lists) == 4
        assert self.utt.label_lists['alpha'] == self.ll_1
        assert self.utt.label_lists['bravo'] == self.ll_2
        assert self.utt.label_lists['charlie'] == self.ll_3
        assert self.utt.label_lists['delta'] == ll_4
        assert self.ll_1.utterance == self.utt
        assert self.ll_2.utterance == self.utt
        assert self.ll_3.utterance == self.utt
        assert ll_4.utterance == self.utt

    def test_label_values(self):
        assert self.utt.all_label_values() == {'a', 'b', 'c', 'd', 'e', 'f', 'g'}

    def test_label_values_with_idx_restriction(self):
        all_values = self.utt.all_label_values(label_list_ids=['bravo', 'charlie'])
        assert all_values == {'a', 'c', 'd', 'e', 'f', 'g'}

    def test_label_count(self):
        assert self.utt.label_count() == {'a': 3, 'b': 1, 'c': 2, 'd': 3, 'e': 1, 'f': 1, 'g': 1}

    def test_label_count_with_idx_restriction(self):
        count = self.utt.label_count(label_list_ids=['bravo', 'charlie'])
        assert count == {'a': 2, 'c': 1, 'd': 1, 'e': 1, 'f': 1, 'g': 1}

    def test_label_duration(self):
        duration = self.utt.label_total_duration(label_list_ids=['alpha', 'charlie'])
        assert duration['a'] == pytest.approx(4.5)
        assert duration['b'] == pytest.approx(3.8)
        assert duration['c'] == pytest.approx(3.3)
        assert duration['d'] == pytest.approx(6.5)
        assert duration['g'] == pytest.approx(3.7)

    def test_read_samples(self):
        expected, __ = librosa.core.load(self.track.path, sr=None, offset=1.25, duration=0.05)
        assert np.array_equal(self.utt.read_samples(), expected)

    def test_read_samples_with_offset(self):
        expected, __ = librosa.core.load(self.track.path, sr=None, offset=1.27, duration=0.03)
        assert np.array_equal(self.utt.read_samples(offset=0.02), expected)

    def test_read_samples_with_duration(self):
        expected, __ = librosa.core.load(self.track.path, sr=None, offset=1.26, duration=0.03)
        assert np.array_equal(self.utt.read_samples(offset=0.01, duration=0.03), expected)

    def test_read_samples_with_duration_no_end_set(self):
        self.utt.end = float('inf')

        expected, __ = librosa.core.load(self.track.path, sr=None, offset=1.26, duration=0.03)
        assert np.array_equal(self.utt.read_samples(offset=0.01, duration=0.03), expected)

    def test_num_samples(self):
        assert self.utt.num_samples() == 800

    def test_num_samples_non_native_sr(self):
        assert self.utt.num_samples(sr=12000) == 600

    def test_num_samples_matches_read_samples(self):
        assert self.utt.read_samples().shape[0] == self.utt.num_samples()
        assert self.utt.read_samples(sr=11255).shape[0] == self.utt.num_samples(sr=11255)

    def test_split(self):
        ll_1 = annotations.LabelList('phones', labels=[
            annotations.Label('alpha', start=0.0, end=30.0)
        ])
        ll_2 = annotations.LabelList('words', labels=[
            annotations.Label('b', start=0.0, end=30.0)
        ])
        utt = tracks.Utterance('utt-1', 'track-x', start=0.0, end=40.0, label_lists=[ll_1, ll_2])

        res = utt.split([14.0, 29.5])

        assert len(res) == 3

        assert res[0].start == 0.0
        assert res[0].end == 14.0
        assert 'phones' in res[0].label_lists.keys()
        assert 'words' in res[0].label_lists.keys()

        assert res[1].start == 14.0
        assert res[1].end == 29.5
        assert 'phones' in res[1].label_lists.keys()
        assert 'words' in res[1].label_lists.keys()

        assert res[2].start == 29.5
        assert res[2].end == 40.0
        assert 'phones' in res[2].label_lists.keys()
        assert 'words' in res[2].label_lists.keys()

    def test_split_endless(self):
        utt = tracks.Utterance('utt-1', None, start=0.0)
        res = utt.split([24.5])

        assert len(res) == 2
        assert res[0].start == 0.0
        assert res[0].end == 24.5
        assert res[1].start == 24.5
        assert res[1].end == float('inf')

    def test_split_sets_track(self):
        file = tracks.FileTrack('file-1', '/some/path')
        utt = tracks.Utterance('utt-1', file, start=0.0, end=10.0)
        res = utt.split([5.2])

        assert len(res) == 2
        assert res[0].track == file
        assert res[1].track == file

    def test_split_sets_issuer(self):
        issuer = issuers.Speaker('spk-1')
        utt = tracks.Utterance('utt-1', None, issuer=issuer, start=0.0, end=10.0)
        res = utt.split([5.2])

        assert len(res) == 2
        assert res[0].issuer == issuer
        assert res[1].issuer == issuer

    def test_split_without_cutting_points_raises_error(self):
        utt = tracks.Utterance('utt-1', None, start=0.0)

        with pytest.raises(ValueError):
            utt.split([])

    def test_split_with_cutting_point_after_end_returns_one_utt(self):
        utt = tracks.Utterance('utt-1', None, start=4.0, end=20.0)
        res = utt.split([24.5])

        assert len(res) == 1
        assert res[0].start == 4.0
        assert res[0].end == 20.0

    def test_split_when_utt_start_is_not_zero(self):
        utt = tracks.Utterance('utt-1', None, start=6.0, end=20.0)
        res = utt.split([3.0])

        assert len(res) == 2
        assert res[0].start == 6.0
        assert res[0].end == 9.0
        assert res[1].start == 9.0
        assert res[1].end == 20.

    def test_split_track_relative(self):
        utt = tracks.Utterance('utt-1', None, start=6.0, end=20.0)
        res = utt.split([8.0], track_relative=True)

        assert len(res) == 2
        assert res[0].start == 6.0
        assert res[0].end == 8.0
        assert res[1].start == 8.0
        assert res[1].end == 20.00

    def test_split_utt_relative(self):
        utt = tracks.Utterance('utt-1', None, start=6.0, end=20.0)
        res = utt.split([8.0], track_relative=False)

        assert len(res) == 2
        assert res[0].start == 6.0
        assert res[0].end == 14.0
        assert res[1].start == 14.0
        assert res[1].end == 20.0

    def test_split_utt_relative_with_labels(self):
        ll_1 = annotations.LabelList('phones', labels=[
            annotations.Label('alpha', start=0.0, end=30.0)
        ])
        ll_2 = annotations.LabelList('words', labels=[
            annotations.Label('b', start=8.0, end=30.0)
        ])
        utt = tracks.Utterance('utt-1', 'file-x', start=10.0, end=40.0, label_lists=[ll_1, ll_2])

        res = utt.split([14.0], track_relative=False)

        assert len(res) == 2

        assert res[0].start == 10.0
        assert res[0].end == 24.0
        assert res[0].label_lists['phones'] == annotations.LabelList(idx='phones', labels=[
            annotations.Label('alpha', 0.0, 14.0)
        ])
        assert res[0].label_lists['words'] == annotations.LabelList(idx='words', labels=[
            annotations.Label('b', 8.0, 14.0)
        ])

        assert res[1].start == 24.0
        assert res[1].end == 40.0
        assert res[1].label_lists['phones'] == annotations.LabelList(idx='phones', labels=[
            annotations.Label('alpha', 0.0, 16.0)
        ])
        assert res[1].label_lists['words'] == annotations.LabelList(idx='words', labels=[
            annotations.Label('b', 0.0, 16.0)
        ])

    def test_split_with_overlap(self):
        ll_1 = annotations.LabelList('phones', labels=[
            annotations.Label('alpha', start=0.0, end=30.0),
            annotations.Label('bravo', start=20.0, end=42.0)
        ])
        ll_2 = annotations.LabelList('words', labels=[
            annotations.Label('b', start=8.0, end=30.0)
        ])
        utt = tracks.Utterance('utt-1', 'file-x', start=10.0, end=55.0, label_lists=[ll_1, ll_2])

        res = utt.split([12.0, 24.0], track_relative=False, overlap=2.0)

        assert len(res) == 3

        assert res[0].start == 10.0
        assert res[0].end == 24.0
        assert res[0].label_lists['phones'] == annotations.LabelList(idx='phones', labels=[
            annotations.Label('alpha', 0.0, 14.0)
        ])
        assert res[0].label_lists['words'] == annotations.LabelList(idx='words', labels=[
            annotations.Label('b', 8.0, 14.0)
        ])

        print(res[1].label_lists['phones'].labels)

        assert res[1].start == 20.0
        assert res[1].end == 36.0
        assert res[1].label_lists['phones'] == annotations.LabelList(idx='phones', labels=[
            annotations.Label('alpha', 0.0, 16.0),
            annotations.Label('bravo', 10.0, 16.0),
        ])
        assert res[1].label_lists['words'] == annotations.LabelList(idx='words', labels=[
            annotations.Label('b', 0.0, 16.0)
        ])

        assert res[2].start == 32.0
        assert res[2].end == 55.0
        assert res[2].label_lists['phones'] == annotations.LabelList(idx='phones', labels=[
            annotations.Label('alpha', 0.0, 8.0),
            annotations.Label('bravo', 0.0, 20.0),
        ])
        assert res[2].label_lists['words'] == annotations.LabelList(idx='words', labels=[
            annotations.Label('b', 0.0, 8.0)
        ])



import os

import numpy as np
from audiomate import encoding

from tests import resources


class EncoderMock(encoding.Encoder):

    def encode_utterance(self, utterance, corpus=None):
        return np.array([1, 2, 3])


class TestEncoder:

    def test_encode_corpus(self, tmpdir):
        ds = resources.create_single_label_corpus()
        target_path = os.path.join(tmpdir.strpath, 'data.hdf5')

        encoder = EncoderMock()
        container = encoder.encode_corpus(ds, target_path)

        with container as ct:
            assert ct.path == target_path
            assert set(ct.keys()) == set(ds.utterances.keys())

            for utterance_idx in ds.utterances.keys():
                assert np.array_equal(ct.get(utterance_idx, mem_map=False), np.array([1, 2, 3]))



import os

from setuptools import find_packages
from setuptools import setup

##################################################
# Dependencies
##################################################

PYTEST_VERSION_ = '4.0.0'

# Packages required in 'production'
REQUIRED = [
    'audioread == 2.1.8',
    'numpy == 1.16.2',
    'scipy == 1.2.1',
    'librosa == 0.6.3',
    'h5py == 2.9.0',
    'networkx == 2.2',
    'requests == 2.21.0',
    'intervaltree == 3.0.2',
]

# Packages required for dev/ci enrionment
EXTRAS = {
    'dev': [
        'click==6.7',
        'pytest==%s' % (PYTEST_VERSION_,),
        'pytest-runner==3.0',
        'pytest-cov==2.5.1',
        'requests_mock==1.4.0',
        'Sphinx==1.8.5',
        'sphinx-rtd-theme==0.4.3',
        'pytest-benchmark==3.1.1',
    ],
    'ci': [
        'flake8==3.6.0',
        'flake8-quotes==0.12.1'
    ],
}

# Packages required for testing
TESTS = [
    'pytest==%s' % (PYTEST_VERSION_,),
    'requests_mock==1.4.0'
]

##################################################
# Description
##################################################

DESCRIPTION = 'Audiomate is a library for working with audio datasets.'

root = os.path.abspath(os.path.dirname(__file__))
readme_path = os.path.join(root, 'README.md')

# Import the README and use it as the long-description.
try:
    with open(readme_path, encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

##################################################
# SETUP
##################################################

setup(name='audiomate',
      version='5.0.0',
      description=DESCRIPTION,
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/ynop/audiomate',
      download_url='https://github.com/ynop/audiomate/releases',
      author='Matthias Buechi, Andreas Ahlenstorf',
      author_email='buec@zhaw.ch',
      classifiers=[
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3 :: Only',
          'Topic :: Scientific/Engineering :: Human Machine Interfaces'
      ],
      keywords='audio music sound corpus dataset',
      license='MIT',
      packages=find_packages(exclude=['tests']),
      install_requires=REQUIRED,
      include_package_data=True,
      zip_safe=False,
      test_suite='tests',
      extras_require=EXTRAS,
      setup_requires=['pytest-runner'],
      tests_require=TESTS,
      entry_points={}
      )



"""
Functions for reading/writing sclite transcription files.

Description of the format:
http://www1.icsi.berkeley.edu/Speech/docs/sctk-1.2/infmts.htm#trn_fmt_name_0
"""

import re

TRANSCRIPT_PATTERN = re.compile(r'(.*) \((.*?)\)')


def write(path, entries):
    """
    Writes an transcription file.

    Args:
        path (str): Path to write the file to.
        entries (dict): List with entries to write.

    Example::

        >>> data = {
        >>>     'utt-1': 'sie',
        >>>     'utt-2': 'hallo',
        >>> }
        >>>
        >>> write_label_file('/some/path/to/file.txt', data)
    """
    sorted_entries = sorted(entries.items(), key=lambda x: x[0])
    lines = ['{} ({})'.format(x[1], x[0]) for x in sorted_entries]

    with open(path, 'w') as f:
        f.write('\n'.join(lines))


def read(path):
    """
    Read the labels from a transcription file.

    Args:
        path (str): Path to the label file.

    Returns:
        dict: Dictionary of transcriptions (utt-idx: transcription)

    Example::

        >>> read_label_file('/path/to/label/file.txt')
        {
            'utt-1': 'sie',
            'utt-2': 'hallo'
        }
    """

    entries = {}

    with open(path, 'r') as f:
        for line in f.readlines():
            line = line.strip()

            m = TRANSCRIPT_PATTERN.match(line)

            if m is not None:
                utt_idx = m.group(2).strip()
                transcription = m.group(1).strip()
                entries[utt_idx] = transcription
            else:
                raise ValueError('Failed to parse line of trn file: {}'.format(line))

    return entries



import numpy as np
import scipy.fftpack as fft
import librosa
from librosa import filters
from librosa import util

from . import base


def stft_from_frames(frames, window='hann', dtype=np.complex64):
    """
    Variation of the librosa.core.stft function,
    that computes the short-time-fourier-transfrom from frames instead from the signal.

    See http://librosa.github.io/librosa/_modules/librosa/core/spectrum.html#stft
    """

    win_length = frames.shape[0]
    n_fft = win_length

    fft_window = filters.get_window(window, win_length, fftbins=True)

    # Reshape so that the window can be broadcast
    fft_window = fft_window.reshape((-1, 1))

    # Pre-allocate the STFT matrix
    stft_matrix = np.empty((int(1 + n_fft // 2), frames.shape[1]),
                           dtype=dtype,
                           order='F')

    # how many columns can we fit within MAX_MEM_BLOCK?
    n_columns = int(util.MAX_MEM_BLOCK / (stft_matrix.shape[0] *
                                          stft_matrix.itemsize))

    for bl_s in range(0, stft_matrix.shape[1], n_columns):
        bl_t = min(bl_s + n_columns, stft_matrix.shape[1])

        # RFFT and Conjugate here to match phase from DPWE code
        stft_matrix[:, bl_s:bl_t] = fft.fft(fft_window *
                                            frames[:, bl_s:bl_t],
                                            axis=0)[:stft_matrix.shape[0]].conj()

    return stft_matrix


class MelSpectrogram(base.Computation):
    """
    Computation step that extracts mel-spectrogram features from the given frames.

    Based on http://librosa.github.io/librosa/generated/librosa.feature.melspectrogram.html

    Args:
        n_mels (int): Number of mel bands to generate.
    """

    def __init__(self, n_mels=128, parent=None, name=None):
        super(MelSpectrogram, self).__init__(parent=parent, name=name)

        self.n_mels = n_mels

    def compute(self, chunk, sampling_rate, corpus=None, utterance=None):
        power_spec = np.abs(stft_from_frames(chunk.data.T)) ** 2
        mel = librosa.feature.melspectrogram(S=power_spec, n_mels=self.n_mels, sr=sampling_rate)

        return mel.T


class MFCC(base.Computation):
    """
    Computation step that extracts mfcc features from the given frames.

    Based on http://librosa.github.io/librosa/generated/librosa.feature.mfcc.html

    Args:
        n_mels (int): Number of mel bands to generate.
        n_mfcc (int): number of MFCCs to return.
    """

    def __init__(self, n_mfcc=13, n_mels=128, parent=None, name=None):
        super(MFCC, self).__init__(parent=parent, name=name)

        self.n_mfcc = n_mfcc
        self.n_mels = n_mels

    def compute(self, chunk, sampling_rate, corpus=None, utterance=None):
        power_spec = np.abs(stft_from_frames(chunk.data.T)) ** 2

        mel = librosa.feature.melspectrogram(S=power_spec, n_mels=self.n_mels, sr=sampling_rate)
        mel_power = librosa.power_to_db(mel)
        mfcc = librosa.feature.mfcc(S=mel_power, n_mfcc=self.n_mfcc)

        return mfcc.T



import numpy as np

from audiomate import tracks
from audiomate import annotations
from audiomate import encoding

import pytest


class TestTokenOrdinalEncoder:

    def test_encode_utterance_with_single_label(self):
        ll = annotations.LabelList(idx='go', labels=[
            annotations.Label('a c b')
        ])
        utt = tracks.Utterance('utt-1', None, label_lists=ll)

        encoder = encoding.TokenOrdinalEncoder('go', ['a', 'b', 'c'])
        encoded = encoder.encode_utterance(utt)

        assert np.array_equal(encoded, [0, 2, 1])

    def test_encode_utterance_with_multiple_non_overlapping_labels(self):
        ll = annotations.LabelList(idx='go', labels=[
            annotations.Label('a c b', start=0, end=5),
            annotations.Label('c b b', start=5, end=9.4),
            annotations.Label('a a a', start=9.5, end=10.2)
        ])
        utt = tracks.Utterance('utt-1', None, label_lists=ll)

        encoder = encoding.TokenOrdinalEncoder('go', ['a', 'b', 'c'])
        encoded = encoder.encode_utterance(utt)

        assert np.array_equal(encoded, [0, 2, 1, 2, 1, 1, 0, 0, 0])

    def test_encode_utterance_with_overlapping_labels_raises_error(self):
        ll = annotations.LabelList(idx='go', labels=[
            annotations.Label('a c b', start=0, end=5),
            annotations.Label('c b b', start=2, end=9.4),
            annotations.Label('a a a', start=9.5, end=10.2)
        ])
        utt = tracks.Utterance('utt-1', None, label_lists=ll)

        encoder = encoding.TokenOrdinalEncoder('go', ['a', 'b', 'c'])

        with pytest.raises(ValueError):
            encoder.encode_utterance(utt)

    def test_encode_utterance_with_missing_token_raises_error(self):
        ll = annotations.LabelList(idx='go', labels=[
            annotations.Label('a c b unknown')
        ])
        utt = tracks.Utterance('utt-1', None, label_lists=ll)

        encoder = encoding.TokenOrdinalEncoder('go', ['a', 'b', 'c'])

        with pytest.raises(ValueError):
            encoder.encode_utterance(utt)

    def test_encode_utterance_with_non_existing_label_list_raises_error(self):
        ll = annotations.LabelList(idx='go', labels=[
            annotations.Label('a c b unknown')
        ])
        utt = tracks.Utterance('utt-1', None, label_lists=ll)

        encoder = encoding.TokenOrdinalEncoder('not_existing', ['a', 'b', 'c'])

        with pytest.raises(ValueError):
            encoder.encode_utterance(utt)

    def test_encode_utterance_with_custom_delimiter(self):
        ll = annotations.LabelList(idx='go', labels=[
            annotations.Label('a, c , b, b')
        ])
        utt = tracks.Utterance('utt-1', None, label_lists=ll)

        encoder = encoding.TokenOrdinalEncoder('go', ['a', 'b', 'c'], token_delimiter=',')
        encoded = encoder.encode_utterance(utt)

        assert np.array_equal(encoded, [0, 2, 1, 1])



import audiomate
from audiomate.corpus import validation

from tests import resources


class TestTrackReadValidator:

    def test_name(self):
        val = validation.TrackReadValidator()
        assert val.name() == 'Track-Read'

    def test_validate_passes(self):
        corpus = audiomate.Corpus()
        corpus.new_file(
            resources.sample_wav_file('wav_1.wav'),
            'wav1'
        )
        corpus.new_file(
            resources.sample_wav_file('wav_2.wav'),
            'wav2'
        )

        val = validation.TrackReadValidator()
        res = val.validate(corpus)

        assert res.passed

    def test_validate_doesnt_pass(self):
        corpus = audiomate.Corpus()
        corpus.new_file(
            resources.sample_wav_file('wav_1.wav'),
            'wav1'
        )
        corpus.new_file(
            resources.sample_wav_file('invalid_audio.wav'),
            'wav2'
        )

        val = validation.TrackReadValidator()
        res = val.validate(corpus)

        assert not res.passed
        assert len(res.invalid_items) == 1
        assert 'wav2' in res.invalid_items



import os

import numpy as np

from audiomate import containers
from audiomate.corpus import subset
from audiomate import feeding
from audiomate.processing import pipeline

import pytest
from tests import resources


@pytest.fixture
def container_dim_x_4(tmpdir):
    inputs_path = os.path.join(tmpdir.strpath, 'inputs.hdf5')

    cnt = containers.FeatureContainer(inputs_path)
    cnt.open()

    cnt.frame_size = 4
    cnt.hop_size = 2
    cnt.sampling_rate = 16000

    cnt.set('utt-1', np.arange(60).reshape(15, 4))
    cnt.set('utt-2', np.arange(80).reshape(20, 4))
    cnt.set('utt-3', np.arange(44).reshape(11, 4))
    cnt.set('utt-4', np.arange(12).reshape(3, 4))
    cnt.set('utt-5', np.arange(16).reshape(4, 4))

    return cnt


@pytest.fixture
def container_dim_x(tmpdir):
    inputs_path = os.path.join(tmpdir.strpath, 'outputs.hdf5')

    cnt = containers.Container(inputs_path)
    cnt.open()

    cnt.set('utt-1', np.arange(6))
    cnt.set('utt-2', np.arange(8))
    cnt.set('utt-3', np.arange(4))
    cnt.set('utt-4', np.arange(2))
    cnt.set('utt-5', np.arange(6))

    return cnt


class TestDataset:

    def test_init_throws_error_when_no_container_is_given(self):
        corpus = resources.create_dataset()

        with pytest.raises(ValueError):
            feeding.Dataset(corpus, [])

    def test_init_with_corpus(self, tmpdir):
        c = containers.Container(os.path.join(tmpdir.strpath, 'test.h5'))
        c.open()
        c.set('utt-1', data=np.arange(20))
        c.set('utt-2', data=np.arange(20))
        c.set('utt-3', data=np.arange(20))
        c.set('utt-4', data=np.arange(20))
        c.set('utt-5', data=np.arange(20))

        corpus = resources.create_dataset()
        it = feeding.Dataset(corpus, [c])
        assert it.utt_ids == ['utt-1', 'utt-2', 'utt-3', 'utt-4', 'utt-5']

    def test_init_with_corpus_view(self, tmpdir):
        c = containers.Container(os.path.join(tmpdir.strpath, 'test.h5'))
        c.open()
        c.set('utt-1', data=np.arange(20))
        c.set('utt-2', data=np.arange(20))
        c.set('utt-3', data=np.arange(20))
        c.set('utt-4', data=np.arange(20))
        c.set('utt-5', data=np.arange(20))

        corpus = resources.create_dataset()
        subview = subset.Subview(corpus, filter_criteria=[
            subset.MatchingUtteranceIdxFilter(utterance_idxs={'utt-1', 'utt-2', 'utt-4'})
        ])

        it = feeding.Dataset(subview, [c])
        assert it.utt_ids == ['utt-1', 'utt-2', 'utt-4']

    def test_init_with_utterance_list(self, tmpdir):
        c = containers.Container(os.path.join(tmpdir.strpath, 'test.h5'))
        c.open()
        c.set('utt-1', data=np.arange(20))
        c.set('utt-2', data=np.arange(20))

        it = feeding.Dataset(['utt-1', 'utt-2'], [c])
        assert it.utt_ids == ['utt-1', 'utt-2']

    def test_init_missing_utterance_in_container_raises_error(self, tmpdir):
        c = containers.Container(os.path.join(tmpdir.strpath, 'test.h5'))
        c.open()
        c.set('utt-1', data=np.arange(20))
        c.set('utt-3', data=np.arange(20))

        with pytest.raises(ValueError):
            feeding.Dataset(['utt-1', 'utt-2', 'utt-3'], [containers.Container('blub')])

    def test_container_has_utterances(self, tmpdir):
        c = containers.Container(os.path.join(tmpdir.strpath, 'test.h5'))
        c.open()
        c.set('utt-1', data=np.arange(20))
        c.set('utt-2', data=np.arange(20))
        c.set('utt-3', data=np.arange(20))

        assert feeding.Dataset.container_has_utterances(c, ['utt-1', 'utt-2', 'utt-3'])

    def test_container_has_utterances_returns_false_if_one_is_missing(self, tmpdir):
        c = containers.Container(os.path.join(tmpdir.strpath, 'test.h5'))
        c.open()
        c.set('utt-1', data=np.arange(20))
        c.set('utt-3', data=np.arange(20))

        assert not feeding.Dataset.container_has_utterances(c, ['utt-1', 'utt-2', 'utt-3'])


class AddTransform(pipeline.Computation):
    def __init__(self, value, parent=None, name=None):
        super(AddTransform, self).__init__(parent=parent, name=name)
        self.value = value

    def compute(self, chunk, sampling_rate, corpus=None, utterance=None):
        return chunk.data + self.value


class TestUtteranceDataset:

    def test_init_with_corpus(self, container_dim_x_4, container_dim_x):
        corpus = resources.create_dataset()
        ds = feeding.UtteranceDataset(corpus, [container_dim_x_4, container_dim_x])

        assert ds.utt_ids == ['utt-1', 'utt-2', 'utt-3', 'utt-4', 'utt-5']

    def test_init_with_list_of_utt_ids(self, container_dim_x_4, container_dim_x):
        ds = feeding.UtteranceDataset(['utt-1', 'utt-2', 'utt-3', 'utt-4', 'utt-5'],
                                      [container_dim_x_4, container_dim_x])

        assert ds.utt_ids == ['utt-1', 'utt-2', 'utt-3', 'utt-4', 'utt-5']

    def test_len(self, container_dim_x_4, container_dim_x):
        ds = feeding.UtteranceDataset(['utt-1', 'utt-2', 'utt-3', 'utt-4', 'utt-5'],
                                      [container_dim_x_4, container_dim_x])

        assert len(ds) == 5

    def test_get_item(self, container_dim_x_4, container_dim_x):
        ds = feeding.UtteranceDataset(['utt-1', 'utt-2', 'utt-3', 'utt-4', 'utt-5'],
                                      [container_dim_x_4, container_dim_x])

        sample = ds[2]

        assert len(sample) == 4
        assert np.array_equal(sample[0], np.arange(44).reshape(11, 4))
        assert sample[1] == 11
        assert np.array_equal(sample[2], np.arange(4))
        assert sample[3] == 4

    def test_get_item_padded(self, container_dim_x_4, container_dim_x):
        ds = feeding.UtteranceDataset(['utt-1', 'utt-2', 'utt-3', 'utt-4', 'utt-5'],
                                      [container_dim_x_4, container_dim_x], pad=True)

        sample = ds[2]

        assert len(sample) == 4
        assert np.array_equal(sample[0], np.pad(np.arange(44).reshape(11, 4), ((0, 9), (0, 0)), mode='constant'))
        assert sample[1] == 11
        assert np.array_equal(sample[2], np.pad(np.arange(4), (0, 4), mode='constant'))
        assert sample[3] == 4

    def test_get_item_padded_when_no_padding_required(self, container_dim_x_4, container_dim_x):
        ds = feeding.UtteranceDataset(['utt-1', 'utt-2', 'utt-3', 'utt-4', 'utt-5'],
                                      [container_dim_x_4, container_dim_x], pad=True)

        sample = ds[1]

        assert len(sample) == 4
        assert np.array_equal(sample[0], np.arange(80).reshape(20, 4))
        assert sample[1] == 20
        assert np.array_equal(sample[2], np.arange(8))
        assert sample[3] == 8

    def test_longest_utterances_per_container(self, container_dim_x_4, container_dim_x):
        ds = feeding.UtteranceDataset(['utt-1', 'utt-2', 'utt-3', 'utt-4', 'utt-5'],
                                      [container_dim_x_4, container_dim_x])

        lengths = ds.longest_utterances_per_container()

        assert len(lengths) == 2
        assert lengths[0] == 20
        assert lengths[1] == 8

    def test_get_item_with_transfrom(self, container_dim_x_4, container_dim_x):
        transform = AddTransform(value=3.0)

        ds = feeding.UtteranceDataset(
            ['utt-1', 'utt-2', 'utt-3', 'utt-4', 'utt-5'],
            [container_dim_x_4, container_dim_x],
            transform=[transform, None]
        )

        sample = ds[2]

        assert len(sample) == 4
        assert np.array_equal(sample[0], np.arange(44).reshape(11, 4) + 3.0)
        assert sample[1] == 11
        assert np.array_equal(sample[2], np.arange(4))
        assert sample[3] == 4

    def test_init_with_transform_for_non_feature_container_raises_error(
            self, container_dim_x_4, container_dim_x):
        transform = AddTransform(value=3.0)

        with pytest.raises(ValueError):
            feeding.UtteranceDataset(
                ['utt-1', 'utt-2', 'utt-3', 'utt-4', 'utt-5'],
                [container_dim_x_4, container_dim_x],
                transform=[transform, transform]
            )

    def test_init_with_missing_transforms_raises_error(self, container_dim_x_4, container_dim_x):
        transform = AddTransform(value=3.0)

        with pytest.raises(ValueError):
            feeding.UtteranceDataset(
                ['utt-1', 'utt-2', 'utt-3', 'utt-4', 'utt-5'],
                [container_dim_x_4, container_dim_x],
                transform=[transform]
            )


@pytest.fixture
def sample_multi_frame_dataset(tmpdir):
    inputs_path = os.path.join(tmpdir.strpath, 'inputs.hdf5')
    targets_path = os.path.join(tmpdir.strpath, 'targets.hdf5')

    corpus = resources.create_dataset()
    container_inputs = containers.Container(inputs_path)
    container_targets = containers.Container(targets_path)

    container_inputs.open()
    container_targets.open()

    container_inputs.set('utt-1', np.arange(60).reshape(15, 4))
    container_inputs.set('utt-2', np.arange(80).reshape(20, 4))
    container_inputs.set('utt-3', np.arange(44).reshape(11, 4))
    container_inputs.set('utt-4', np.arange(12).reshape(3, 4))
    container_inputs.set('utt-5', np.arange(16).reshape(4, 4))

    container_targets.set('utt-1', np.arange(30).reshape(15, 2))
    container_targets.set('utt-2', np.arange(40).reshape(20, 2))
    container_targets.set('utt-3', np.arange(22).reshape(11, 2))
    container_targets.set('utt-4', np.arange(6).reshape(3, 2))
    container_targets.set('utt-5', np.arange(8).reshape(4, 2))

    return feeding.MultiFrameDataset(corpus, [container_inputs, container_targets], 4)


class TestMultiFrameDataset:

    def test_raises_error_if_frames_per_chunk_is_smaller_than_one(self):
        with pytest.raises(ValueError):
            feeding.MultiFrameDataset(['utt-1'], [], 0)

    def test_partitioned_iterator(self, sample_multi_frame_dataset):
        it = sample_multi_frame_dataset.partitioned_iterator('960', shuffle=True, seed=12)

        assert isinstance(it, feeding.MultiFrameIterator)
        assert it.return_length == sample_multi_frame_dataset.return_length
        assert it.pad == sample_multi_frame_dataset.pad
        assert it.frames_per_chunk == sample_multi_frame_dataset.frames_per_chunk
        assert it.containers == sample_multi_frame_dataset.containers
        assert it.utt_ids == sample_multi_frame_dataset.utt_ids
        assert it.shuffle
        assert it.partition_size == '960'

    def test_get_utt_regions(self, sample_multi_frame_dataset):
        regions = sample_multi_frame_dataset.get_utt_regions()

        assert len(regions) == 5

        assert regions[0][0] == 0
        assert regions[0][1] == 4

        assert regions[1][0] == 4
        assert regions[1][1] == 5

        assert regions[2][0] == 9
        assert regions[2][1] == 3

        assert regions[3][0] == 12
        assert regions[3][1] == 1

        assert regions[4][0] == 13
        assert regions[4][1] == 1

    def test_get_length(self, sample_multi_frame_dataset):
        assert len(sample_multi_frame_dataset) == 14

    def test_get_item_in_the_middle_of_an_utterance(self, sample_multi_frame_dataset):
        assert len(sample_multi_frame_dataset[2]) == 2
        assert np.array_equal(sample_multi_frame_dataset[2][0], np.arange(16).reshape(4, 4) + 32)
        assert np.array_equal(sample_multi_frame_dataset[2][1], np.arange(8).reshape(4, 2) + 16)

    def test_get_item_at_the_begin_of_an_utterance(self, sample_multi_frame_dataset):
        assert len(sample_multi_frame_dataset[4]) == 2
        assert np.array_equal(sample_multi_frame_dataset[4][0], np.arange(16).reshape(4, 4))
        assert np.array_equal(sample_multi_frame_dataset[4][1], np.arange(8).reshape(4, 2))

    def test_get_item_at_the_end_of_an_utterance(self, sample_multi_frame_dataset):
        assert len(sample_multi_frame_dataset[11]) == 2
        assert np.array_equal(sample_multi_frame_dataset[11][0], np.arange(12).reshape(3, 4) + 32)
        assert np.array_equal(sample_multi_frame_dataset[11][1], np.arange(6).reshape(3, 2) + 16)

    def test_return_correct_length_for_chunk_with_full_size(self, sample_multi_frame_dataset):
        ds_length_enabled = feeding.MultiFrameDataset(sample_multi_frame_dataset.utt_ids,
                                                      sample_multi_frame_dataset.containers,
                                                      4,
                                                      return_length=True)

        assert len(ds_length_enabled[9]) == 3
        assert ds_length_enabled[9][2] == 4

    def test_return_correct_length_for_chunk_at_end_of_utterance(self, sample_multi_frame_dataset):
        ds_length_enabled = feeding.MultiFrameDataset(sample_multi_frame_dataset.utt_ids,
                                                      sample_multi_frame_dataset.containers,
                                                      4,
                                                      return_length=True)

        assert len(ds_length_enabled[11]) == 3
        assert ds_length_enabled[11][2] == 3

    def test_pads_shorter_chunks_with_zeros(self, sample_multi_frame_dataset):
        ds_pad_enabled = feeding.MultiFrameDataset(sample_multi_frame_dataset.utt_ids,
                                                   sample_multi_frame_dataset.containers,
                                                   4,
                                                   pad=True)

        assert len(ds_pad_enabled[11]) == 3
        exp = np.pad(np.arange(12).reshape(3, 4) + 32, ((0, 1), (0, 0)), mode='constant', constant_values=0)
        assert np.array_equal(ds_pad_enabled[11][0], exp)
        exp = np.pad(np.arange(6).reshape(3, 2) + 16, ((0, 1), (0, 0)), mode='constant', constant_values=0)
        assert np.array_equal(ds_pad_enabled[11][1], exp)
        assert ds_pad_enabled[11][2] == 3


@pytest.fixture
def sample_frame_dataset(tmpdir):
    inputs_path = os.path.join(tmpdir.strpath, 'inputs.hdf5')
    targets_path = os.path.join(tmpdir.strpath, 'targets.hdf5')

    corpus = resources.create_dataset()
    container_inputs = containers.Container(inputs_path)
    container_targets = containers.Container(targets_path)

    container_inputs.open()
    container_targets.open()

    container_inputs.set('utt-1', np.arange(20).reshape(5, 4))
    container_inputs.set('utt-2', np.arange(28).reshape(7, 4))
    container_inputs.set('utt-3', np.arange(36).reshape(9, 4))
    container_inputs.set('utt-4', np.arange(8).reshape(2, 4))
    container_inputs.set('utt-5', np.arange(16).reshape(4, 4))

    container_targets.set('utt-1', np.arange(20).reshape(5, 4) + 10)
    container_targets.set('utt-2', np.arange(28).reshape(7, 4) + 10)
    container_targets.set('utt-3', np.arange(36).reshape(9, 4) + 10)
    container_targets.set('utt-4', np.arange(8).reshape(2, 4) + 10)
    container_targets.set('utt-5', np.arange(16).reshape(4, 4) + 10)

    return feeding.FrameDataset(corpus, [container_inputs, container_targets])


class TestFrameDataset:

    def test_partitioned_iterator(self, sample_frame_dataset):
        it = sample_frame_dataset.partitioned_iterator('960', shuffle=True, seed=12)

        assert isinstance(it, feeding.FrameIterator)
        assert it.containers == sample_frame_dataset.containers
        assert it.utt_ids == sample_frame_dataset.utt_ids
        assert it.shuffle
        assert it.partition_size == '960'

    def test_get_length(self, sample_frame_dataset):
        assert len(sample_frame_dataset) == 27

    def test_get_item(self, sample_frame_dataset):
        assert len(sample_frame_dataset[4]) == 2
        assert np.array_equal(sample_frame_dataset[4][0], np.array([16, 17, 18, 19]))
        assert np.array_equal(sample_frame_dataset[4][1], np.array([16, 17, 18, 19]) + 10)

        assert len(sample_frame_dataset[12]) == 2
        assert np.array_equal(sample_frame_dataset[12][0], np.array([0, 1, 2, 3]))
        assert np.array_equal(sample_frame_dataset[12][1], np.array([0, 1, 2, 3]) + 10)

        assert len(sample_frame_dataset[22]) == 2
        assert np.array_equal(sample_frame_dataset[22][0], np.array([4, 5, 6, 7]))
        assert np.array_equal(sample_frame_dataset[22][1], np.array([4, 5, 6, 7]) + 10)

        assert len(sample_frame_dataset[26]) == 2
        assert np.array_equal(sample_frame_dataset[26][0], np.array([12, 13, 14, 15]))
        assert np.array_equal(sample_frame_dataset[26][1], np.array([12, 13, 14, 15]) + 10)



import librosa
from audiomate.utils import audioread
import numpy as np
import scipy


def process_buffer(buffer, n_channels):
    """
    Merge the read blocks and resample if necessary.

    Args:
        buffer (list): A list of blocks of samples.
        n_channels (int): The number of channels of the input data.

    Returns:
        np.array: The samples
    """
    samples = np.concatenate(buffer)

    if n_channels > 1:
        samples = samples.reshape((-1, n_channels)).T
        samples = librosa.to_mono(samples)

    return samples


def read_blocks(file_path, start=0.0, end=float('inf'), buffer_size=5760000):
    """
    Read an audio file block after block. The blocks are yielded one by one.

    Args:
        file_path (str): Path to the file to read.
        start (float): Start in seconds to read from.
        end (float): End in seconds to read to.
                     ``inf`` means to the end of the file.
        buffer_size (int): Number of samples to load into memory at once and
                           return as a single block. The exact number of loaded
                           samples depends on the block-size of the
                           audioread library. So it can be of x higher,
                           where the x is typically 1024 or 4096.

    Returns:
        Generator: A generator yielding the samples for every block.
    """
    buffer = []
    n_buffer = 0
    n_samples = 0

    with audioread.audio_open(file_path) as input_file:
        n_channels = input_file.channels
        sr_native = input_file.samplerate

        start_sample = int(np.round(sr_native * start)) * n_channels
        end_sample = end

        if end_sample != np.inf:
            end_sample = int(np.round(sr_native * end)) * n_channels

        for block in input_file:
            block = librosa.util.buf_to_float(block)
            n_prev = n_samples
            n_samples += len(block)

            if n_samples < start_sample:
                continue

            if n_prev > end_sample:
                break

            if n_samples > end_sample:
                block = block[:end_sample - n_prev]

            if n_prev <= start_sample <= n_samples:
                block = block[start_sample - n_prev:]

            n_buffer += len(block)
            buffer.append(block)

            if n_buffer >= buffer_size:
                yield process_buffer(buffer, n_channels)

                buffer = []
                n_buffer = 0

        if len(buffer) > 0:
            yield process_buffer(buffer, n_channels)


def read_frames(file_path, frame_size, hop_size, start=0.0,
                end=float('inf'), buffer_size=5760000):
    """
    Read an audio file frame by frame. The frames are yielded one after another.

    Args:
        file_path (str): Path to the file to read.
        frame_size (int): The number of samples per frame.
        hop_size (int): The number of samples between two frames.
        start (float): Start in seconds to read from.
        end (float): End in seconds to read to.
                     ``inf`` means to the end of the file.
        buffer_size (int): Number of samples to load into memory at once
                           and return as a single block.
                           The exact number of loaded samples depends on the
                           block-size of the audioread library. So it can be
                           of x higher, where the x is typically 1024 or 4096.

    Returns:
        Generator: A generator yielding a tuple for every frame.
        The first item is the frame and
        the second a boolean indicating if it is the last frame.
    """
    rest_samples = np.array([], dtype=np.float32)

    for block in read_blocks(file_path, start=start, end=end, buffer_size=buffer_size):

        # Prepend rest samples from previous block
        block = np.concatenate([rest_samples, block])

        current_sample = 0

        # Get frames that are fully contained in the block
        while current_sample + frame_size < block.size:
            frame = block[current_sample:current_sample + frame_size]
            yield frame, False
            current_sample += hop_size

        # Store rest samples for next block
        rest_samples = block[current_sample:]

    if rest_samples.size > 0:
        rest_samples = np.pad(
            rest_samples,
            (0, frame_size - rest_samples.size),
            mode='constant',
            constant_values=0
        )
        yield rest_samples, True


def write_wav(path, samples, sr=16000):
    """
    Write to given samples to a wav file.
    The samples are expected to be floating point numbers
    in the range of -1.0 to 1.0.

    Args:
        path (str): The path to write the wav to.
        samples (np.array): A float array .
        sr (int): The sampling rate.
    """
    max_value = np.abs(np.iinfo(np.int16).min)
    data = (samples * max_value).astype(np.int16)
    scipy.io.wavfile.write(path, sr, data)



import math

from . import base


class MeanVarianceNorm(base.Computation):
    """
    Pre-processing step to normalize mean and variance.

    frame = (frame - mean) / sqrt(variance)

    Args:
        mean (float): The mean to use for normalization.
        variance (float): The variance to use for normalization.s
    """

    def __init__(self, mean, variance, parent=None, name=None):
        super(MeanVarianceNorm, self).__init__(parent=parent, name=name)

        self.mean = mean
        self.variance = variance
        self.std = math.sqrt(variance)

    def compute(self, chunk, sampling_rate, corpus=None, utterance=None):
        return (chunk.data - self.mean) / self.std



import collections
import random

from . import subview
from . import utils


class SubsetGenerator(object):
    """
    This class is used to generate subsets of a corpus.

    Args:
        corpus (Corpus): The corpus to create subsets from.
        random_seed (int): Seed to use for random number generation.
    """

    def __init__(self, corpus, random_seed=None):
        self.corpus = corpus
        self.random_seed = random_seed
        self.rand = random.Random()
        self.rand.seed(a=random_seed)

    def random_subset(self, relative_size, balance_labels=False, label_list_ids=None):
        """
        Create a subview of random utterances with a approximate size relative to the full corpus.
        By default x random utterances are selected with x equal to ``relative_size * corpus.num_utterances``.

        Args:
            relative_size (float): A value between 0 and 1.
                                   (0.5 will create a subset with approximately 50% of the full corpus size)
            balance_labels (bool): If True, the labels of the selected utterances are balanced as far as possible.
                                   So the count/duration of every label within the subset is equal.
            label_list_ids (list): List of label-list ids. If none is given, all label-lists are considered
                                   for balancing. Otherwise only the ones that are in the list are considered.

        Returns:
            Subview: The subview representing the subset.
        """

        num_utterances_in_subset = round(relative_size * self.corpus.num_utterances)
        all_utterance_ids = sorted(list(self.corpus.utterances.keys()))

        if balance_labels:
            all_label_values = self.corpus.all_label_values(label_list_ids=label_list_ids)
            utterance_with_label_counts = collections.defaultdict(dict)

            for utterance_idx, utterance in self.corpus.utterances.items():
                utterance_with_label_counts[utterance_idx] = utterance.label_count(label_list_ids=label_list_ids)

            subset_utterance_ids = utils.select_balanced_subset(utterance_with_label_counts,
                                                                num_utterances_in_subset,
                                                                list(all_label_values),
                                                                seed=self.rand.random())

        else:
            subset_utterance_ids = self.rand.sample(all_utterance_ids,
                                                    num_utterances_in_subset)

        filter = subview.MatchingUtteranceIdxFilter(utterance_idxs=set(subset_utterance_ids))
        return subview.Subview(self.corpus, filter_criteria=[filter])

    def random_subset_by_duration(self, relative_duration, balance_labels=False, label_list_ids=None):
        """
        Create a subview of random utterances with a approximate duration relative to the full corpus.
        Random utterances are selected so that the sum of all utterance durations
        equals to the relative duration of the full corpus.

        Args:
            relative_duration (float): A value between 0 and 1. (e.g. 0.5 will create a subset with approximately
                                       50% of the full corpus duration)
            balance_labels (bool): If True, the labels of the selected utterances are balanced as far as possible.
                                   So the count/duration of every label within the subset is equal.
            label_list_ids (list): List of label-list ids. If none is given, all label-lists are considered
                                   for balancing. Otherwise only the ones that are in the list are considered.

        Returns:
            Subview: The subview representing the subset.
        """
        total_duration = self.corpus.total_duration
        subset_duration = relative_duration * total_duration
        utterance_durations = {utt_idx: utt.duration for utt_idx, utt in self.corpus.utterances.items()}

        if balance_labels:
            all_label_values = self.corpus.all_label_values(label_list_ids=label_list_ids)

            label_durations = {}
            for utt_idx, utt in self.corpus.utterances.items():
                label_durations[utt_idx] = utt.label_total_duration(label_list_ids)

            subset_utterance_ids = utils.select_balanced_subset(label_durations,
                                                                subset_duration,
                                                                list(all_label_values),
                                                                select_count_values=utterance_durations,
                                                                seed=self.rand.random())

        else:
            dummy_weights = {utt_idx: {'w': 1} for utt_idx in self.corpus.utterances.keys()}
            subset_utterance_ids = utils.select_balanced_subset(dummy_weights,
                                                                subset_duration,
                                                                ['w'],
                                                                select_count_values=utterance_durations,
                                                                seed=self.rand.random())

        filter = subview.MatchingUtteranceIdxFilter(utterance_idxs=set(subset_utterance_ids))
        return subview.Subview(self.corpus, filter_criteria=[filter])

    def random_subsets(self, relative_sizes, by_duration=False, balance_labels=False, label_list_ids=None):
        """
        Create a bunch of subsets with the given sizes relative to the size or duration of the full corpus.
        Basically the same as calling ``random_subset`` or ``random_subset_by_duration`` multiple times
        with different values. But this method makes sure that every subset contains only utterances,
        that are also contained in the next bigger subset.

        Args:
            relative_sizes (list): A list of numbers between 0 and 1 indicating the sizes of the desired subsets,
                                   relative to the full corpus.
            by_duration (bool): If True the size measure is the duration of all utterances in a subset/corpus.
            balance_labels (bool): If True the labels contained in a subset are chosen to be balanced
                                   as far as possible.
            label_list_ids (list): List of label-list ids. If none is given, all label-lists are considered
                                   for balancing. Otherwise only the ones that are in the list are considered.

        Returns:
            dict : A dictionary containing all subsets with the relative size as key.
        """
        resulting_sets = {}
        next_bigger_subset = self.corpus

        for relative_size in reversed(relative_sizes):
            generator = SubsetGenerator(next_bigger_subset, random_seed=self.random_seed)

            if by_duration:
                sv = generator.random_subset_by_duration(relative_size, balance_labels=balance_labels,
                                                         label_list_ids=label_list_ids)
            else:
                sv = generator.random_subset(relative_size, balance_labels=balance_labels,
                                             label_list_ids=label_list_ids)

            resulting_sets[relative_size] = sv

        return resulting_sets

    def maximal_balanced_subset(self, by_duration=False, label_list_ids=None):
        """
        Create a subset of the corpus as big as possible, so that the labels are balanced approximately.
        The label with the shortest duration (or with the fewest utterance if by_duration=False) is taken as reference.
        All other labels are selected so they match the shortest one as far as possible.

        Args:
            by_duration (bool): If True the size measure is the duration of all utterances in a subset/corpus.
            label_list_ids (list): List of label-list ids. If none is given, all label-lists are considered
                                   for balancing. Otherwise only the ones that are in the list are considered.

        Returns:
            Subview: The subview representing the subset.
        """
        all_label_values = self.corpus.all_label_values(label_list_ids=label_list_ids)

        if by_duration:
            utterance_durations = {utt_idx: utt.duration for utt_idx, utt in self.corpus.utterances.items()}
            total_duration_per_label = self.corpus.label_durations(label_list_ids=label_list_ids)
            rarest_label_duration = sorted(total_duration_per_label.values())[0]
            target_duration = len(all_label_values) * rarest_label_duration

            label_durations_per_utterance = {}
            for utt_idx, utt in self.corpus.utterances.items():
                label_durations_per_utterance[utt_idx] = utt.label_total_duration(label_list_ids)

            subset_utterance_ids = utils.select_balanced_subset(label_durations_per_utterance,
                                                                target_duration,
                                                                list(all_label_values),
                                                                select_count_values=utterance_durations,
                                                                seed=self.rand.random())

        else:
            total_count_per_label = self.corpus.label_count(label_list_ids=label_list_ids)
            lowest_label_count = sorted(total_count_per_label.values())[0]
            target_label_count = lowest_label_count * len(all_label_values)
            utterance_with_label_counts = collections.defaultdict(dict)

            for utterance_idx, utterance in self.corpus.utterances.items():
                utterance_with_label_counts[utterance_idx] = utterance.label_count(label_list_ids=label_list_ids)

            subset_utterance_ids = utils.select_balanced_subset(utterance_with_label_counts,
                                                                target_label_count,
                                                                list(all_label_values),
                                                                seed=self.rand.random())

        filter = subview.MatchingUtteranceIdxFilter(utterance_idxs=set(subset_utterance_ids))
        return subview.Subview(self.corpus, filter_criteria=[filter])



import os
import shutil

import audiomate
from audiomate import annotations
from audiomate import issuers
from audiomate.utils import download
from audiomate.utils import textfile
from . import base

SENTENCE_LIST_URL = 'http://downloads.tatoeba.org/exports/sentences.tar.bz2'
AUDIO_LIST_URL = 'http://downloads.tatoeba.org/exports/sentences_with_audio.tar.bz2'

META_FILENAME = 'meta.txt'


class TatoebaDownloader(base.CorpusDownloader):
    """
    Downloader for audio files from the tatoeba platform.

    .. seealso::

       `Tatoeba <https://tatoeba.org/>`_
          Website

    Args:
        include_languages (list): List of languages to download. If None all are downloaded.
        include_licenses (list): Sentences are downloaded only if their license is in this list.
                                 If None all licenses are included.
        load_empty_license (bool): Sentences with an empty license are not meant to be reused.
                                   If False these sentences are ignored.
    """

    def __init__(self, include_languages=None, include_licenses=None, include_empty_licence=False):
        self.include_languages = include_languages
        self.include_licenses = include_licenses
        self.include_empty_licence = include_empty_licence

    @classmethod
    def type(cls):
        return 'tatoeba'

    def _download(self, target_path):
        temp_path = os.path.join(target_path, 'temp')
        os.makedirs(temp_path, exist_ok=True)

        sentence_ark = os.path.join(temp_path, 'sentences.tar.bz2')
        sentence_list = os.path.join(temp_path, 'sentences.csv')
        audio_ark = os.path.join(temp_path, 'sentences_with_audio.tar.bz2')
        audio_list = os.path.join(temp_path, 'sentences_with_audio.csv')

        download.download_file(SENTENCE_LIST_URL, sentence_ark)
        download.download_file(AUDIO_LIST_URL, audio_ark)

        download.extract_tar(sentence_ark, temp_path)
        download.extract_tar(audio_ark, temp_path)

        audio_entries = self._load_audio_list(audio_list)
        sentences = self._load_sentence_list(sentence_list)

        valid_sentence_ids = set(audio_entries.keys()).intersection(set(sentences.keys()))

        # sent-id, username, lang, transcript
        all_records = [(k, audio_entries[k][0], sentences[k][0], sentences[k][1]) for k in valid_sentence_ids]

        meta_path = os.path.join(target_path, META_FILENAME)
        textfile.write_separated_lines(meta_path, all_records, separator='\t', sort_by_column=0)

        self._download_audio_files(all_records, target_path)

        shutil.rmtree(temp_path, ignore_errors=True)

    def _load_audio_list(self, path):
        """
        Load and filter the audio list.

        Args:
            path (str): Path to the audio list file.

        Returns:
            dict: Dictionary of filtered sentences (id : username, license, attribution-url)
        """

        result = {}

        for entry in textfile.read_separated_lines_generator(path, separator='\t', max_columns=4):
            for i in range(len(entry)):
                if entry[i] == '\\N':
                    entry[i] = None

            if len(entry) < 4:
                entry.extend([None] * (4 - len(entry)))

            if not self.include_empty_licence and entry[2] is None:
                continue

            if self.include_licenses is not None and entry[2] not in self.include_licenses:
                continue

            result[entry[0]] = entry[1:]

        return result

    def _load_sentence_list(self, path):
        """
        Load and filter the sentence list.

        Args:
            path (str): Path to the sentence list.

        Returns:
            dict: Dictionary of sentences (id : language, transcription)
        """

        result = {}

        for entry in textfile.read_separated_lines_generator(path, separator='\t', max_columns=3):
            if self.include_languages is None or entry[1] in self.include_languages:
                result[entry[0]] = entry[1:]

        return result

    def _download_audio_files(self, records, target_path):
        """
        Download all audio files based on the given records.
        """

        for record in records:
            audio_folder = os.path.join(target_path, 'audio', record[2])
            audio_file = os.path.join(audio_folder, '{}.mp3'.format(record[0]))
            os.makedirs(audio_folder, exist_ok=True)

            download_url = 'https://audio.tatoeba.org/sentences/{}/{}.mp3'.format(record[2], record[0])
            download.download_file(download_url, audio_file)


class TatoebaReader(base.CorpusReader):
    """
    Reader for audio data downloaded with the Tatoeba downloader.
    """

    @classmethod
    def type(cls):
        return 'tatoeba'

    def _check_for_missing_files(self, path):
        meta_file = os.path.join(path, META_FILENAME)

        if os.path.isfile(meta_file):
            return []
        else:
            return [meta_file]

    def _load(self, path):
        corpus = audiomate.Corpus(path=path)

        meta_file = os.path.join(path, META_FILENAME)
        records = textfile.read_separated_lines_generator(meta_file, separator='\t', max_columns=4)

        for record in records:
            idx = record[0]
            speaker_idx = record[1]
            language = record[2]
            transcript = record[3]

            file_path = os.path.join(path, 'audio', language, '{}.mp3'.format(idx))
            corpus.new_file(file_path, idx)

            if speaker_idx not in corpus.issuers.keys():
                issuer = issuers.Speaker(speaker_idx)
                corpus.import_issuers(issuer)

            utterance = corpus.new_utterance(idx, idx, speaker_idx)
            utterance.set_label_list(annotations.LabelList.create_single(transcript,
                                                                         idx=audiomate.corpus.LL_WORD_TRANSCRIPT_RAW))

        return corpus



import numpy as np
import librosa

from audiomate.processing import pipeline


class TestMelSpectrogram:

    def test_compute(self):
        samples = np.arange(8096).astype(np.float32)
        D = np.abs(librosa.core.stft(samples, n_fft=2048, hop_length=512, center=False)) ** 2
        expected = librosa.feature.melspectrogram(S=D, sr=16000, n_mels=128).T

        frames = librosa.util.frame(samples, frame_length=2048, hop_length=512).T
        mel = pipeline.MelSpectrogram(n_mels=128)
        res = mel.process_frames(frames, sampling_rate=16000)

        assert np.array_equal(expected, res)


class TestMFCC:

    def test_compute(self):
        samples = np.arange(8096).astype(np.float32)
        D = np.abs(librosa.core.stft(samples, n_fft=2048, hop_length=512, center=False)) ** 2
        mel = librosa.feature.melspectrogram(S=D, sr=16000, n_mels=128)
        expected = librosa.feature.mfcc(S=librosa.power_to_db(mel), n_mfcc=13).T

        frames = librosa.util.frame(samples, frame_length=2048, hop_length=512).T
        mfcc = pipeline.MFCC(n_mfcc=13, n_mels=128)
        res = mfcc.process_frames(frames, sampling_rate=16000)

        assert np.array_equal(expected, res)



import os
from collections import namedtuple

from audiomate import tracks
from audiomate import issuers
from audiomate import annotations

import pytest


ExpFileTrack = namedtuple(
    'FileTrack',
    ['idx', 'path']
)
ExpContainerTrack = namedtuple(
    'ContainerTrack',
    ['idx', 'key', 'container_path']
)

ExpIssuer = namedtuple(
    'Issuer',
    ['idx', 'num_utterances']
)

ExpSpeaker = namedtuple(
    'Speaker',
    ['idx', 'num_utterances', 'gender', 'age_group', 'native_lang']
)

ExpArtist = namedtuple(
    'Artist',
    ['idx', 'num_utterances', 'name']
)

ExpUtterance = namedtuple(
    'Utterance',
    ['idx', 'track_idx', 'issuer_idx', 'start', 'end']
)

ExpLabelList = namedtuple(
    'LabelList',
    ['ll_idx', 'num_labels']
)

ExpLabel = namedtuple(
    'Label',
    ['ll_idx', 'value', 'start', 'end']
)

ExpSubview = namedtuple(
    'Subview',
    ['idx', 'utterance_ids']
)


class CorpusReaderTest:
    """
    Base class for testing a corpus reader.

    All the basic components can be tested by defining the expected assets
    using the EXPECTED-variables.
    """

    FILE_TRACK_BASE_PATH = None

    EXPECTED_NUMBER_OF_TRACKS = None
    EXPECTED_TRACKS = []

    EXPECTED_NUMBER_OF_UTTERANCES = None
    EXPECTED_UTTERANCES = []

    EXPECTED_NUMBER_OF_ISSUERS = None
    EXPECTED_ISSUERS = []

    # Expected Label-Lists/Labels per utterance
    EXPECTED_LABEL_LISTS = {}
    EXPECTED_LABELS = {}

    EXPECTED_NUMBER_OF_SUBVIEWS = None
    EXPECTED_SUBVIEWS = []

    def load(self):
        pass

    # ----------------------------------------------------------------------
    # TRACKS
    # ----------------------------------------------------------------------

    def test_number_of_tracks(self):
        assert self.EXPECTED_NUMBER_OF_TRACKS is not None, \
            'Expected number of tracks has to be defined'

        corpus = self.load()

        assert corpus.num_tracks == self.EXPECTED_NUMBER_OF_TRACKS, \
            'Number of tracks in corpus is wrong'

    def test_tracks(self):
        if len(self.EXPECTED_TRACKS) == 0:
            pytest.skip('No tracks expected')

        corpus = self.load()

        for exp in self.EXPECTED_TRACKS:
            idx = exp.idx

            assert idx in corpus.tracks.keys(), \
                'track with idx "{}" is missing'.format(idx)

            actual = corpus.tracks[idx]

            assert actual.idx == idx, \
                'track idx "{}" differs from track-dictionary idx "{}"'.format(
                    idx, actual.idx)

            if isinstance(exp, ExpFileTrack):
                assert isinstance(actual, tracks.FileTrack), \
                    'Track "{}" has a wrong type'.format(idx)

                exp_path = exp.path

                if self.FILE_TRACK_BASE_PATH is not None:
                    exp_path = os.path.join(
                        self.FILE_TRACK_BASE_PATH,
                        exp.path
                    )
                assert exp_path == actual.path, \
                    'Path of file "{}" is wrong'.format(idx)
            elif isinstance(exp, ExpFileTrack):
                assert isinstance(actual, tracks.ContainerTrack), \
                    'Track "{}" has a wrong type'.format(idx)
                assert actual.key == exp.key, \
                    'Key of container-track "{}" is wrong'.format(idx)
                assert actual.container.path == exp.key, \
                    'Container-path of track "{}" is wrong'.format(idx)

    # ----------------------------------------------------------------------
    # UTTERANCES
    # ----------------------------------------------------------------------

    def test_number_of_utterances(self):
        assert self.EXPECTED_NUMBER_OF_UTTERANCES is not None, \
            'Expected number of utterances has to be defined'

        corpus = self.load()

        assert corpus.num_utterances == self.EXPECTED_NUMBER_OF_UTTERANCES, \
            'Number of utterances in corpus is wrong'

    def test_utterances(self):
        if len(self.EXPECTED_UTTERANCES) == 0:
            pytest.skip('No utterances expected')

        corpus = self.load()

        for exp in self.EXPECTED_UTTERANCES:
            idx = exp.idx

            assert idx in corpus.utterances.keys(), \
                'Utterance with idx "{}" is missing'.format(idx)
            assert corpus.utterances[idx].idx == idx, \
                ('utterance idx "{}" differs from'
                 'utterance-dictionary idx "{}"').format(
                idx, corpus.tracks[idx].idx
            )

            actual = corpus.utterances[idx]
            assert actual.start == exp.start, \
                'Start of utterance is wrong'
            assert actual.end == exp.end, \
                'End of utterance is wrong'

            if exp.issuer_idx is None:
                assert actual.issuer is None, \
                    'Utterance is expected to have no issuer'
            else:
                assert actual.issuer.idx == exp.issuer_idx, \
                    'Issuer-idx of utterance {} is wrong'.format(idx)

    # ----------------------------------------------------------------------
    # ISSUERS
    # ----------------------------------------------------------------------

    def test_number_of_issuers(self):
        assert self.EXPECTED_NUMBER_OF_ISSUERS is not None, \
            'Expected number of issuers has to be defined'

        corpus = self.load()

        assert corpus.num_issuers == self.EXPECTED_NUMBER_OF_ISSUERS, \
            'Number of issuers in corpus is wrong'

    def test_issuers(self):
        corpus = self.load()

        for exp in self.EXPECTED_ISSUERS:
            idx = exp.idx

            assert idx in corpus.issuers.keys(), \
                'Issuer with idx "{}" is missing'.format(idx)

            actual = corpus.issuers[idx]

            assert actual.idx == idx, \
                ('Issuer idx "{}" differs from'
                 'issuer-dictionary idx "{}"').format(
                    idx, actual.idx
                )
            assert len(actual.utterances) == exp.num_utterances, \
                'Issuer "{}" has the wrong num. of utterances'.format(idx)

            if isinstance(exp, ExpSpeaker):
                assert isinstance(actual, issuers.Speaker), \
                    'Issuer "{}" has a wrong type'.format(idx)
                assert actual.gender == exp.gender, \
                    'Gender of speaker "{}" is wrong'.format(idx)
                assert actual.age_group == exp.age_group, \
                    'AgeGroup of speaker "{}" is wrong'.format(idx)
                assert actual.native_language == exp.native_lang, \
                    'Native language of speaker "{}" is wrong'.format(idx)
            elif isinstance(exp, ExpArtist):
                assert isinstance(actual, issuers.Artist), \
                    'Issuer "{}" has a wrong type'.format(idx)
                assert actual.name == exp.name, \
                    'Name of artist "{}" is wrong'.format(idx)
            elif isinstance(exp, ExpIssuer):
                assert isinstance(actual, issuers.Issuer), \
                    'Issuer "{}" has a wrong type'.format(idx)

    # ----------------------------------------------------------------------
    # LABEL / LABELLISTS
    # ----------------------------------------------------------------------

    def test_label_lists(self):
        corpus = self.load()

        for utt_idx, utt_label_lists in self.EXPECTED_LABEL_LISTS.items():
            utt = corpus.utterances[utt_idx]

            for exp_ll in utt_label_lists:
                idx = exp_ll.ll_idx

                assert idx in utt.label_lists.keys(), \
                    'LabelList with idx "{}" is missing'.format(idx)

                actual_ll = utt.label_lists[idx]

                assert actual_ll.idx == idx, \
                    ('LabelList idx "{}" differs from'
                     'labellists-dictionary idx "{}"').format(
                        idx, actual_ll.idx
                    )
                assert len(actual_ll) == exp_ll.num_labels, \
                    '[{}] Number of labels in label-list "{}" is wrong'.format(utt_idx, idx)

    def test_labels(self):
        corpus = self.load()

        for utt_idx, utt_labels in self.EXPECTED_LABELS.items():
            utt = corpus.utterances[utt_idx]

            for exp in utt_labels:
                ll = utt.label_lists[exp.ll_idx]
                expected = annotations.Label(exp.value, exp.start, exp.end)

                assert expected in list(ll), \
                    '{} not in LabelList {}'.format(expected, ll.idx)

    # ----------------------------------------------------------------------
    # SUBVIEWS
    # ----------------------------------------------------------------------

    def test_number_of_subviews(self):
        assert self.EXPECTED_NUMBER_OF_SUBVIEWS is not None, \
            'Expected number of subviews has to be defined'

        corpus = self.load()

        assert corpus.num_subviews == self.EXPECTED_NUMBER_OF_SUBVIEWS, \
            'Number of subviews in corpus is wrong'

    def test_subviews(self):
        corpus = self.load()

        for exp in self.EXPECTED_SUBVIEWS:
            idx = exp.idx

            assert idx in corpus.subviews.keys(), \
                'Subview with idx "{}" is missing'.format(idx)

            actual = corpus.subviews[idx]

            assert list(sorted(actual.utterances.keys())) == \
                list(sorted(exp.utterance_ids)), \
                'Subviews has mismatches of utterance ids'



import os
import collections

import audiomate
from audiomate import annotations
from audiomate.corpus import subset
from . import base
from audiomate.utils import textfile


class Urbansound8kReader(base.CorpusReader):
    """
    Reader for the Urbansound8k dataset.

    .. seealso::

       `Urbansound8k <http://urbansounddataset.weebly.com/urbansound8k.html>`_
          Download page
    """

    @classmethod
    def type(cls):
        return 'urbansound8k'

    def _check_for_missing_files(self, path):
        meta_file_path = os.path.join(path, 'metadata', 'UrbanSound8K.csv')

        if os.path.isfile(meta_file_path):
            return []
        else:
            return [meta_file_path]

    def _load(self, path):
        corpus = audiomate.Corpus(path=path)

        meta_file_path = os.path.join(path, 'metadata', 'UrbanSound8K.csv')
        meta_data = textfile.read_separated_lines(meta_file_path, separator=',', max_columns=8)[1:]

        folds = collections.defaultdict(set)

        for record in meta_data:
            file_name = record[0]
            fold = record[5]
            label = record[7]

            file_path = os.path.join(path, 'audio', 'fold{}'.format(fold), file_name)
            if os.path.isfile(file_path):
                basename = os.path.splitext(file_name)[0]

                corpus.new_file(file_path, basename)
                utt = corpus.new_utterance(basename, basename)
                utt.set_label_list(annotations.LabelList.create_single(label, idx=audiomate.corpus.LL_SOUND_CLASS))
                folds['fold{}'.format(fold)].add(basename)

        for fold_idx, fold_utterance_ids in folds.items():
            filter = subset.MatchingUtteranceIdxFilter(utterance_idxs=fold_utterance_ids)
            subview = subset.Subview(corpus, filter_criteria=[filter])

            corpus.import_subview(fold_idx, subview)

        return corpus



import os
import shutil


def move_all_files_from_subfolders_to_top(folder_path, delete_subfolders=False, copy=False):
    """
    Move all files/folder from all subfolders of `folder_path` on top into `folder_path`.

    Args:
        folder_path (str): Path of the folder.
        delete_subfolders (bool): If True the subfolders are deleted after all items are moved out of it.
        copy (bool): If True copies the files instead of moving. (default False)
    """
    for item in os.listdir(folder_path):
        sub_path = os.path.join(folder_path, item)

        if os.path.isdir(sub_path):

            for sub_item in os.listdir(sub_path):
                src = os.path.join(sub_path, sub_item)
                target = os.path.join(folder_path, sub_item)

                if copy:
                    if os.path.isfile(src):
                        shutil.copy(src, target)
                    else:
                        shutil.copytree(src, target)
                else:
                    shutil.move(src, target)

            if delete_subfolders:
                shutil.rmtree(sub_path)



import numpy as np

from audiomate.processing import pipeline


class TestAvgPool:

    def test_compute(self):
        avg_pooling = pipeline.AvgPool(3)

        data_a = np.array([
            [1, 1],
            [2, 2],
            [3, 3],
            [4, 4],
            [5, 5]
        ])
        chunk_a = pipeline.Chunk(data_a, 0, False)

        out_a = avg_pooling.compute(chunk_a, 16000)

        assert np.allclose(out_a, np.array([
            [2, 2]
        ]))

        data_b = np.array([
            [6, 1],
            [7, 1],
            [8, 1],
            [9, 1],
        ])
        chunk_b = pipeline.Chunk(data_b, 5, True)

        out_b = avg_pooling.compute(chunk_b, 16000)

        assert np.allclose(out_b, np.array([
            [5, 10.0 / 3.0],
            [8, 1]
        ]))

    def test_compute_with_rest(self):
        avg_pooling = pipeline.AvgPool(3)

        data_a = np.array([
            [1, 1],
            [2, 2],
            [3, 3],
            [4, 4],
            [5, 5]
        ])
        chunk_a = pipeline.Chunk(data_a, 0, True)

        out_a = avg_pooling.compute(chunk_a, 16000)

        assert np.allclose(out_a, np.array([
            [2, 2],
            [4.5, 4.5]
        ]))

    def test_compute_cleanup_after_one_utterance(self):
        avg_pooling = pipeline.AvgPool(3)

        data_a = np.array([
            [1, 1],
            [2, 2],
            [3, 3],
            [4, 4],
            [5, 5]
        ])
        chunk_a = pipeline.Chunk(data_a, 0, True)

        out_a = avg_pooling.compute(chunk_a, 16000)

        assert np.allclose(out_a, np.array([
            [2, 2],
            [4.5, 4.5]
        ]))

        out_a = avg_pooling.compute(chunk_a, 16000)

        assert np.allclose(out_a, np.array([
            [2, 2],
            [4.5, 4.5]
        ]))

    def test_frame_transform_step(self):
        avg_pooling = pipeline.AvgPool(3)

        fs, hs = avg_pooling.frame_transform(400, 160)

        assert fs == 400 + 2 * 160
        assert hs == 3 * 160

    def test_frame_transform_step_full_hop(self):
        avg_pooling = pipeline.AvgPool(3)

        fs, hs = avg_pooling.frame_transform(400, 400)

        assert fs == 3 * 400
        assert hs == 3 * 400


class TestVarPool:

    def test_compute(self):
        var_pooling = pipeline.VarPool(3)

        data_a = np.array([
            [1, 1],
            [2, 2],
            [3, 3],
            [4, 4],
            [5, 5]
        ])
        chunk_a = pipeline.Chunk(data_a, 0, False)

        out_a = var_pooling.compute(chunk_a, 16000)

        assert np.allclose(out_a, np.array([
            [0.6666666666666666, 0.6666666666666666]
        ]))

        data_b = np.array([
            [6, 1],
            [7, 1],
            [8, 1],
            [9, 1],
        ])
        chunk_b = pipeline.Chunk(data_b, 5, True)

        out_b = var_pooling.compute(chunk_b, 16000)

        assert np.allclose(out_b, np.array([
            [0.6666666666666666, 2.8888888888888893],
            [0.6666666666666666, 0]
        ]))

    def test_compute_with_rest(self):
        var_pooling = pipeline.VarPool(3)

        data_a = np.array([
            [1, 1],
            [2, 2],
            [3, 3],
            [4, 4],
            [5, 5]
        ])
        chunk_a = pipeline.Chunk(data_a, 0, True)

        out_a = var_pooling.compute(chunk_a, 16000)

        assert np.allclose(out_a, np.array([
            [0.6666666666666666, 0.6666666666666666],
            [0.25, 0.25]
        ]))

    def test_compute_cleanup_after_one_utterance(self):
        var_pooling = pipeline.VarPool(3)

        # FIRST
        data_a = np.array([
            [1, 1],
            [2, 2],
            [3, 3],
            [4, 4],
            [5, 5]
        ])
        chunk_a = pipeline.Chunk(data_a, 0, True)

        out_a = var_pooling.compute(chunk_a, 16000)

        assert np.allclose(out_a, np.array([
            [0.6666666666666666, 0.6666666666666666],
            [0.25, 0.25]
        ]))

        # SECOND
        out_a = var_pooling.compute(chunk_a, 16000)

        assert np.allclose(out_a, np.array([
            [0.6666666666666666, 0.6666666666666666],
            [0.25, 0.25]
        ]))

    def test_frame_transform_step(self):
        var_pooling = pipeline.VarPool(10)

        fs, hs = var_pooling.frame_transform(400, 160)

        assert fs == 400 + 9 * 160
        assert hs == 10 * 160



"""
This module contains functions for validating a corpus on different properties.
e.g. if the length of the utterance is to short for its corresponding transcription.

:py:class:`audiomate.corpus.validation.Validator` is the base class for performing validations.
It can be extended to implement validators for specific tests/validations.
Thre result of every validator has to be a :py:class:`audiomate.corpus.validation.ValidationResult`
or a subclass of it.
"""

from .base import Validator  # noqa: F401
from .base import ValidationResult  # noqa: F401
from .base import InvalidItemsResult  # noqa: F401

from .combine import CombinedValidator  # noqa: F401
from .combine import CombinedValidationResult  # noqa: F401

from .label_list import UtteranceTranscriptionRatioValidator  # noqa: F401
from .label_list import LabelCountValidator  # noqa: F401
from .label_list import LabelCoverageValidator  # noqa: F401
from .label_list import LabelCoverageValidationResult  # noqa: F401
from .label_list import LabelOverflowValidator  # noqa: F401
from .label_list import LabelOverflowValidationResult  # noqa: F401

from .track import TrackReadValidator  # noqa: F401



import copy
from functools import total_ordering


@total_ordering
class Label(object):
    """
    Represents a label that describes some part of an utterance.

    Args:
        value (str): The text of the label.
        start (float): Start of the label within the utterance in seconds.
                       (default: 0)
        end (float): End of the label within the utterance in seconds.
                     (default: inf) (inf defines the end of the utterance)
        meta (dict): A dictionary containing additional information
                     for the label.

    Attributes:
        label_list (LabelList): The label-list this label is belonging to.
    """
    __slots__ = ['value', 'start', 'end', 'label_list', 'meta']

    def __init__(self, value, start=0, end=float('inf'), meta=None):
        self.value = value
        self.start = start
        self.end = end
        self.meta = meta or {}
        self.label_list = None

    def __eq__(self, other):
        return (
            self.start == other.start and
            self.end == other.end and
            self.value.lower() == other.value.lower()
        )

    def __lt__(self, other):
        data_this = (self.start, self.end, self.value.lower())
        data_other = (other.start, other.end, other.value.lower())

        return data_this < data_other

    def __repr__(self) -> str:
        return 'Label({}, {}, {})'.format(self.value, self.start, self.end)

    def __copy__(self):
        return Label(
            self.value,
            start=self.start,
            end=self.end,
            meta=self.meta
        )

    def __deepcopy__(self, memo):
        return Label(
            self.value,
            start=self.start,
            end=self.end,
            meta=copy.deepcopy(self.meta, memo)
        )

    @property
    def start_abs(self):
        """
        Return the absolute start of the label in seconds relative to
        the signal. If the label isn't linked to any utterance via label-list,
        it is assumed ``self.start`` is relative to the start of the signal,
        hence ``self.start`` == ``self.start_abs``.
        """
        if self.label_list is None or self.label_list.utterance is None:
            return self.start

        return self.label_list.utterance.start + self.start

    @property
    def end_abs(self):
        """
        Return the absolute end of the label in seconds relative to the signal.
        If the label isn't linked to any utterance via label-list,
        it is assumed ``self.end`` is relative to the start of the signal,
        hence ``self.end`` == ``self.end_abs``.
        """
        if self.label_list is None or self.label_list.utterance is None:
            return self.end
        elif self.end == float('inf'):
            return self.label_list.utterance.end_abs
        else:
            return self.end + self.label_list.utterance.start

    @property
    def duration(self):
        """
        Return the duration of the label in seconds.
        """
        return self.end_abs - self.start_abs

    @property
    def length(self):
        """
        Return the length of the label (Number of characters).
        """
        return len(self.value)

    def read_samples(self, sr=None):
        """
        Read the samples of the utterance.

        Args:
            sr (int): If None uses the sampling rate given by the track,
                      otherwise resamples to the given sampling rate.

        Returns:
            np.ndarray: A numpy array containing the samples as a
            floating point (numpy.float32) time series.
        """
        duration = None

        if self.end != float('inf') or self.label_list.utterance.end >= 0:
            duration = self.duration

        track = self.label_list.utterance.track

        return track.read_samples(
            sr=sr,
            offset=self.start_abs,
            duration=duration
        )

    def tokenized(self, delimiter=' '):
        """
        Return a list with tokens from the value of the label.
        Tokens are extracted by splitting the string using ``delimiter`` and
        then trimming any whitespace before and after splitted strings.

        Args:
            delimiter (str): The delimiter used to split into tokens.
                             (default: space)

        Return:
            list: A list of tokens in the order they occur in the label.

        Examples:

            >>> label = Label('as is oh')
            >>> label.tokenized()
            ['as', 'is', 'oh']

        Using a different delimiter (whitespace is trimmed anyway):

            >>> label = Label('oh hi, as, is  ')
            >>> label.tokenized(delimiter=',')
            ['oh hi', 'as', 'is']
        """

        tokens = self.value.split(sep=delimiter)
        tokens = [t.strip() for t in tokens]

        while '' in tokens:
            tokens.remove('')

        return tokens

    def do_overlap(self, other_label, adjacent=True):
        """
        Determine whether ``other_label`` overlaps with this label.
        If ``adjacent==True``, adjacent labels are also considered as overlapping.

        Args:
            other_label (Label): Another label.
            adjacent (bool): If ``True``, adjacent labels are
                             considered as overlapping.

        Returns:
            bool: ``True`` if the two labels overlap, ``False`` otherwise.
        """
        this_end = self.end
        other_end = other_label.end

        if this_end == float('inf'):
            this_end = self.end_abs

        if other_end == float('inf'):
            other_end = other_label.end_abs

        if adjacent:
            first_ends_before = this_end < other_label.start
            second_ends_before = other_end < self.start
        else:
            first_ends_before = this_end <= other_label.start
            second_ends_before = other_end <= self.start

        return not (first_ends_before or second_ends_before)

    def overlap_duration(self, other_label):
        """
        Return the duration of the overlapping part between this label
        and ``other_label``.

        Args:
            other_label(Label): Another label to check.

        Return:
            float: The duration of overlap in seconds.

        Example:
            >>> label_a = Label('a', 3.4, 5.6)
            >>> label_b = Label('b', 4.8, 6.2)
            >>> label_a.overlap_duration(label_b)
            0.8
        """
        this_end = self.end
        other_end = other_label.end

        if this_end == float('inf'):
            this_end = self.end_abs

        if other_end == float('inf'):
            other_end = other_label.end_abs

        start_overlap = max(self.start, other_label.start)
        end_overlap = min(this_end, other_end)

        return max(0, end_overlap - start_overlap)



"""
The encoding module provides functionality to encode labels to use for example for training a DNN.
"""

from .base import Encoder  # noqa: F401

from .frame_based import FrameHotEncoder  # noqa: F401
from .frame_based import FrameOrdinalEncoder  # noqa: F401

from .utterance_based import TokenOrdinalEncoder  # noqa: F401



"""
The processing module provides tools for processing audio data in a batch-wise manner.
The idea is to setup a predefined tool that can process all the audio from a corpus.

The basic component is the :py:class:`audiomate.processing.Processor`. It provides the functionality
to reduce any input component like a corpus, feature-container, utterance, file to the abstraction of frames.
A concrete implementation then only has to provide the proper method to process these frames.

Often in audio processing the same components are used in combination with others.
For this purpose a pipeline can be built that processes the frames in multiple steps.
The :mod:`audiomate.processing.pipeline` provides the :py:class:`audiomate.processing.pipeline.Computation`
and :py:class:`audiomate.processing.pipeline.Reduction` classes.
These abstract classes can be extended to create processing components of a pipeline.
The different components are then be coupled to create custom pipelines.
"""

from .base import Processor  # noqa: F401



import copy

import librosa

from . import track
from audiomate.utils import audio
from audiomate.utils import audioread


class FileTrack(track.Track):
    """
    A track that is stored in a file.

    Args:
        idx (str): A identifier to uniquely identify a track.
        path (str): The path to the file.
    """
    __slots__ = ['path']

    def __init__(self, idx, path):
        super(FileTrack, self).__init__(idx)
        self.path = path

    def __copy__(self):
        return FileTrack(self.idx, self.path)

    def __deepcopy(self, memo):
        return copy.copy(self)

    @property
    def sampling_rate(self):
        """
        Return the sampling rate.
        """
        with audioread.audio_open(self.path) as f:
            return f.samplerate

    @property
    def num_channels(self):
        """
        Return the number of channels.
        """
        with audioread.audio_open(self.path) as f:
            return f.channels

    @property
    def num_samples(self):
        """
        Return the total number of samples.
        """
        with audioread.audio_open(self.path) as f:
            return int(f.duration * f.samplerate)

    @property
    def duration(self):
        """
        Return the duration in seconds.
        """
        with audioread.audio_open(self.path) as f:
            return f.duration

    def read_samples(self, sr=None, offset=0, duration=None):
        """
        Return the samples from the file.
        Uses librosa for loading
        (see http://librosa.github.io/librosa/generated/librosa.core.load.html).

        Args:
            sr (int): If ``None``, uses the sampling rate given by the file,
                      otherwise resamples to the given sampling rate.
            offset (float): The time in seconds, from where to start reading
                            the samples (rel. to the file start).
            duration (float): The length of the samples to read in seconds.

        Returns:
            np.ndarray: A numpy array containing the samples as a
            floating point (numpy.float32) time series.
        """
        samples, __ = librosa.core.load(
            self.path,
            sr=sr,
            offset=offset,
            duration=duration
        )
        return samples

    def read_frames(self, frame_size, hop_size, offset=0,
                    duration=None, buffer_size=5760000):
        """
        Generator that reads and returns the samples of the track in frames.

        Args:
            frame_size (int): The number of samples per frame.
            hop_size (int): The number of samples between two frames.
            offset (float): The time in seconds, from where to start
                            reading the samples (rel. to the track start).
            duration (float): The length of the samples to read in seconds.

        Returns:
            Generator: A generator yielding a tuple for every frame.
            The first item is the frame,
            the second the sampling-rate and
            the third a boolean indicating if it is the last frame.
        """
        if duration is not None:
            end = offset + duration
        else:
            end = float('inf')

        return audio.read_frames(
            self.path,
            frame_size,
            hop_size,
            start=offset,
            end=end,
            buffer_size=buffer_size)



import numpy as np
import librosa

from audiomate import tracks
from audiomate.processing import pipeline

from tests import resources


class TestTempogram:

    def test_compute(self):
        test_file_path = resources.sample_wav_file('wav_1.wav')
        y, sr = librosa.load(test_file_path, sr=None)
        frames = librosa.util.frame(y, frame_length=2048, hop_length=1024).T

        # EXPECTED
        S = np.abs(librosa.stft(y, center=False, n_fft=2048, hop_length=1024)) ** 2
        S = librosa.feature.melspectrogram(S=S, n_mels=128, sr=sr)
        S = librosa.power_to_db(S)
        onsets = librosa.onset.onset_strength(S=S, center=False)
        exp_tgram = librosa.feature.tempogram(onset_envelope=onsets, sr=sr, win_length=11, center=True).T

        # ACTUAL
        tgram_step = pipeline.Tempogram(win_length=11)
        tgrams = tgram_step.process_frames(frames, sr, last=True)

        assert np.allclose(tgrams, exp_tgram)

    def test_compute_online(self):
        # Data: 41523 samples, 16 kHz
        # yields 40 frames with frame-size 2048 and hop-size 1024
        test_file_path = resources.sample_wav_file('wav_1.wav')
        y, sr = librosa.load(test_file_path, sr=None)

        # EXPECTED
        y_pad = np.pad(y, (0, 1024), mode='constant', constant_values=0)
        S = np.abs(librosa.stft(y_pad, center=False, n_fft=2048, hop_length=1024)) ** 2
        S = librosa.feature.melspectrogram(S=S, n_mels=128, sr=sr)
        S = librosa.power_to_db(S)
        onsets = librosa.onset.onset_strength(S=S, center=False)
        exp_tgram = librosa.feature.tempogram(onset_envelope=onsets, sr=sr, win_length=4, center=True).T

        # ACTUAL
        test_file = tracks.FileTrack('idx', test_file_path)
        tgram_step = pipeline.Tempogram(win_length=4)
        tgram_gen = tgram_step.process_track_online(test_file, 2048, 1024, chunk_size=5)

        chunks = list(tgram_gen)
        tgrams = np.vstack(chunks)

        assert np.allclose(tgrams, exp_tgram)

    def test_compute_cleanup_after_one_utterance(self):
        test_file_path = resources.sample_wav_file('wav_1.wav')
        y, sr = librosa.load(test_file_path, sr=None)
        frames = librosa.util.frame(y, frame_length=2048, hop_length=1024).T

        # EXPECTED
        S = np.abs(librosa.stft(y, center=False, n_fft=2048, hop_length=1024)) ** 2
        S = librosa.feature.melspectrogram(S=S, n_mels=128, sr=sr)
        S = librosa.power_to_db(S)
        onsets = librosa.onset.onset_strength(S=S, center=False)
        exp_tgram = librosa.feature.tempogram(onset_envelope=onsets, sr=sr, win_length=11, center=True).T

        # ACTUAL
        tgram_step = pipeline.Tempogram(win_length=11)

        # FIRST RUN
        tgrams = tgram_step.process_frames(frames, sr, last=True)

        assert np.allclose(tgrams, exp_tgram)

        # SECOND RUN
        tgrams = tgram_step.process_frames(frames, sr, last=True)

        assert np.allclose(tgrams, exp_tgram)



from audiomate.corpus import validation

from tests import resources


def create_mock_validator(name, passed):
    class MockValidator(validation.Validator):

        def __init__(self, passed):
            self.passed = passed
            self.called = None

        def name(self):
            return name

        def validate(self, corpus):
            self.called = corpus
            return validation.ValidationResult(self.passed, name=self.name())

    return MockValidator(passed)


class TestCombinedValidator:

    def test_validate_does_not_pass(self):
        cv = validation.CombinedValidator(validators=[
            create_mock_validator('a', True),
            create_mock_validator('b', True),
            create_mock_validator('c', False),
            create_mock_validator('d', True)
        ])

        test_corpus = resources.create_dataset()
        result = cv.validate(test_corpus)

        assert not result.passed
        assert len(result.results) == 4

        assert result.results['a'].passed
        assert result.results['b'].passed
        assert not result.results['c'].passed
        assert result.results['d'].passed

        for i in range(4):
            assert cv.validators[i].called == test_corpus


class TestCombinedValidationResult:

    def test_get_report(self):
        result = validation.CombinedValidationResult(passed=False,
                                                     results={
                                                         'a': validation.ValidationResult(passed=True, name='a'),
                                                         'c': validation.ValidationResult(passed=False, name='c'),
                                                         'b': validation.ValidationResult(passed=True, name='b'),
                                                         'd': validation.ValidationResult(passed=True, name='d')
                                                     })

        assert result.get_report() == '\n'.join([
            'a --> Passed',
            'b --> Passed',
            'c --> Failed',
            'd --> Passed',
            '',
            '',
            'a',
            '=',
            '',
            'Result: Passed',
            '',
            '',
            'b',
            '=',
            '',
            'Result: Passed',
            '',
            '',
            'c',
            '=',
            '',
            'Result: Failed',
            '',
            '',
            'd',
            '=',
            '',
            'Result: Passed',
            '',
            ''
        ])



import os
import glob

import audiomate
from . import base


class FolderReader(base.CorpusReader):
    """
    Loads all wavs from the given folder and creates a corpus from it.
    """

    @classmethod
    def type(cls):
        return 'folder'

    def _check_for_missing_files(self, path):
        return []

    def _load(self, path):
        corpus = audiomate.Corpus(path=path)

        for item in glob.glob('{}/*.wav'.format(path), recursive=True):
            basename, __ = os.path.splitext(os.path.basename(item))
            corpus.new_file(item, basename)
            corpus.new_utterance(basename, basename)

        return corpus



import os
import struct

import numpy as np
import scipy

import audiomate
from audiomate import tracks
from audiomate import annotations
from audiomate import issuers
from audiomate.utils import textfile
from . import base
from . import default

WAV_FILE_NAME = 'wav.scp'
SEGMENTS_FILE_NAME = 'segments'
UTT2SPK_FILE_NAME = 'utt2spk'
SPK2GENDER_FILE_NAME = 'spk2gender'
TRANSCRIPTION_FILE_NAME = 'text'
FEATS_FILE_NAME = 'feats'


class KaldiReader(base.CorpusReader):
    """
    Supports reading data sets in Kaldi format.

    .. seealso::

       `Kaldi: Data preparation <http://kaldi-asr.org/doc/data_prep.html>`_
          Describes how a data set has to be structured to be understood
          by Kaldi and the format of the individual files.
    """

    def __init__(self, main_label_list_idx=audiomate.corpus.LL_WORD_TRANSCRIPT,
                 main_feature_idx='default'):
        self.main_label_list_idx = main_label_list_idx
        self.main_feature_idx = main_feature_idx

    @classmethod
    def type(cls):
        return 'kaldi'

    def _check_for_missing_files(self, path):
        necessary_files = [WAV_FILE_NAME, TRANSCRIPTION_FILE_NAME]
        missing_files = []

        for file_name in necessary_files:
            file_path = os.path.join(path, file_name)

            if not os.path.isfile(file_path):
                missing_files.append(file_name)

        return missing_files

    def _load(self, path):
        wav_file_path = os.path.join(path, WAV_FILE_NAME)
        spk2gender_path = os.path.join(path, SPK2GENDER_FILE_NAME)
        utt2spk_path = os.path.join(path, UTT2SPK_FILE_NAME)
        segments_path = os.path.join(path, SEGMENTS_FILE_NAME)
        text_path = os.path.join(path, TRANSCRIPTION_FILE_NAME)

        corpus = audiomate.Corpus(path=path)

        default.DefaultReader.read_files(wav_file_path, corpus)
        KaldiReader.read_genders(spk2gender_path, corpus)
        utt2spk = default.DefaultReader.read_utt_to_issuer_mapping(utt2spk_path, corpus)
        KaldiReader.read_utterances(segments_path, corpus, utt2spk)
        KaldiReader.read_transcriptions(text_path, corpus)

        return corpus

    @staticmethod
    def read_genders(genders_path, corpus):
        if os.path.isfile(genders_path):
            speakers = textfile.read_key_value_lines(genders_path, separator=' ')

            for speaker_idx, gender_str in speakers.items():
                if gender_str == 'm':
                    gender = issuers.Gender.MALE
                else:
                    gender = issuers.Gender.FEMALE

                speaker = issuers.Speaker(speaker_idx, gender=gender)
                corpus.import_issuers(speaker)

    @staticmethod
    def read_utterances(segments_path, corpus, utt2spk):
        # load utterances
        if os.path.isfile(segments_path):
            utterances = textfile.read_separated_lines_with_first_key(
                segments_path,
                separator=' ',
                max_columns=4
            )

            for utt_id, utt_info in utterances.items():
                start = 0
                end = float('inf')

                if len(utt_info) > 1:
                    start = float(utt_info[1])

                if len(utt_info) > 2:
                    end = float(utt_info[2])

                    if end == -1:
                        end = float('inf')

                speaker_idx = None

                if utt_id in utt2spk.keys():
                    speaker_idx = utt2spk[utt_id].idx

                corpus.new_utterance(
                    utt_id, utt_info[0],
                    issuer_idx=speaker_idx,
                    start=start,
                    end=end
                )
        else:
            for file_idx in corpus.files.keys():
                speaker_idx = None

                if file_idx in utt2spk.keys():
                    speaker_idx = utt2spk[file_idx].idx

                corpus.new_utterance(file_idx, file_idx, issuer_idx=speaker_idx)

    @staticmethod
    def read_transcriptions(text_path, corpus):
        transcriptions = textfile.read_key_value_lines(text_path, separator=' ')
        for utt_id, transcription in transcriptions.items():
            ll = annotations.LabelList.create_single(
                transcription,
                idx=audiomate.corpus.LL_WORD_TRANSCRIPT
            )
            corpus.utterances[utt_id].set_label_list(ll)


class KaldiWriter(base.CorpusWriter):
    """
    Supports writing data sets in Kaldi format.

    Args:
        main_label_list_idx (str): The idx of the label-list to use
                                   for writing to transcriptions file.
        main_feature_idx (str): The idx of the feature-container to export.
        use_utt_idx_if_no_speaker_available (bool): If ``True``, the
                                                    utterance-idx is used as
                                                    speaker-idx in the utt2spk
                                                    file, if no speaker exists
                                                    for an utterance.
        create_spk2gender (bool): If ``True`` creates the file spk2gender.
        default_gender (str): If ``create_spk2gender==True`` and the gender of
                              an issuer is not known,
                              this default value will be used (default 'm').
        prefix_utterances_with_speaker (bool): If ``True``, add a prefix in
                                               form of the issuer-idx to
                                               every utterance.
        use_absolute_times (bool): If ``True``, doesn't use -1 for segment ends,
                                   but reads the audio to get absolute duration.

    .. seealso::

       `Kaldi: Data preparation <http://kaldi-asr.org/doc/data_prep.html>`_
          Describes how a data set has to be structured to be
          understood by Kaldi and the format of the individual files.
    """

    def __init__(self, main_label_list_idx=audiomate.corpus.LL_WORD_TRANSCRIPT,
                 main_feature_idx='default', use_utt_idx_if_no_speaker_available=True,
                 create_spk2gender=False, default_gender='m',
                 prefix_utterances_with_speaker=True, use_absolute_times=False):
        self.main_label_list_idx = main_label_list_idx
        self.main_feature_idx = main_feature_idx
        self.use_utt_idx_if_no_speaker_available = use_utt_idx_if_no_speaker_available
        self.create_spk2gender = create_spk2gender
        self.default_gender = default_gender
        self.prefix_utterances_with_speaker = prefix_utterances_with_speaker
        self.use_absolute_times = use_absolute_times

    @classmethod
    def type(cls):
        return 'kaldi'

    def _save(self, corpus, path):
        wav_file_path = os.path.join(path, WAV_FILE_NAME)
        spk2gender_path = os.path.join(path, SPK2GENDER_FILE_NAME)
        utt2spk_path = os.path.join(path, UTT2SPK_FILE_NAME)
        segments_path = os.path.join(path, SEGMENTS_FILE_NAME)
        text_path = os.path.join(path, TRANSCRIPTION_FILE_NAME)

        KaldiWriter.write_tracks(wav_file_path, corpus, path)
        self._write_segments(segments_path, corpus)
        self._write_utt_to_issuer_mapping(utt2spk_path, corpus)
        self._write_transcriptions(text_path, corpus)
        self._write_features(path, corpus)

        if self.create_spk2gender:
            self._write_genders(spk2gender_path, corpus)

    @staticmethod
    def write_tracks(file_path, corpus, path):
        file_records = []

        export_path = os.path.join(path, 'audio')

        for track in corpus.tracks.values():
            if isinstance(track, tracks.FileTrack):
                file_records.append([
                    track.idx,
                    KaldiWriter.extended_filename(track)
                ])

            elif isinstance(track, tracks.ContainerTrack):
                if not os.path.isdir(export_path):
                    os.makedirs(export_path)

                target_path = os.path.join(
                    export_path,
                    '{}.wav'.format(track.idx)
                )

                max_value = np.iinfo(np.int16).max
                samples = (track.read_samples() * max_value).astype(np.int16)
                sampling_rate = track.sampling_rate
                scipy.io.wavfile.write(target_path, sampling_rate, samples)

                file_records.append([
                    track.idx,
                    target_path
                ])

        textfile.write_separated_lines(
            file_path,
            file_records,
            separator=' ',
            sort_by_column=0
        )

    @staticmethod
    def extended_filename(file_track):
        """
        Create extended filename.
        Kaldi only supports wav.
        Therefore other files have to be converted using sox.
        """
        ext = os.path.splitext(file_track.path)[1]
        abs_path = os.path.abspath(file_track.path)

        if ext == '.wav':
            return abs_path
        else:
            return 'sox {} -t wav - |'.format(abs_path)

    def _write_segments(self, utterance_path, corpus):
        utterances = corpus.utterances.values()
        utterance_records = {}

        for u in utterances:
            utt_idx = self._get_utt_idx(u)
            track_idx = u.track.idx
            start = u.start
            end = u.end

            if end == float('inf'):
                if self.use_absolute_times:
                    end = u.end_abs
                else:
                    end = -1

            utterance_records[utt_idx] = [track_idx, start, end]

        textfile.write_separated_lines(
            utterance_path,
            utterance_records,
            separator=' ',
            sort_by_column=0
        )

    def _write_genders(self, gender_path, corpus):
        genders = {}

        for issuer in corpus.issuers.values():
            gender = self.default_gender

            if type(issuer) == issuers.Speaker:
                if issuer.gender == issuers.Gender.MALE:
                    gender = 'm'
                elif issuer.gender == issuers.Gender.FEMALE:
                    gender = 'f'

            genders[issuer.idx] = gender

        if len(genders) > 0:
            textfile.write_separated_lines(
                gender_path,
                genders,
                separator=' ',
                sort_by_column=0
            )

    def _write_transcriptions(self, text_path, corpus):
        transcriptions = {}

        for utterance in corpus.utterances.values():
            utt_idx = self._get_utt_idx(utterance)

            if self.main_label_list_idx in utterance.label_lists.keys():
                label_list = utterance.label_lists[self.main_label_list_idx]
                transcriptions[utt_idx] = ' '.join([l.value for l in label_list])

        textfile.write_separated_lines(
            text_path,
            transcriptions,
            separator=' ',
            sort_by_column=0
        )

    def _write_utt_to_issuer_mapping(self, utt_issuer_path, corpus):
        utt_issuer_records = {}

        for utterance in corpus.utterances.values():
            utt_idx = self._get_utt_idx(utterance)
            if utterance.issuer is not None:
                utt_issuer_records[utt_idx] = utterance.issuer.idx
            elif self.use_utt_idx_if_no_speaker_available:
                utt_issuer_records[utt_idx] = utt_idx

        textfile.write_separated_lines(
            utt_issuer_path,
            utt_issuer_records,
            separator=' ',
            sort_by_column=0
        )

    def _write_features(self, path, corpus):
        if self.main_feature_idx in corpus.feature_containers.keys():
            fc = corpus.features_containers[self.main_feature_idx]
            fc.open()
            matrices = {}

            for utt_id in corpus.utterances.keys():
                matrix = fc.get(utt_id)

                if matrix is not None:
                    matrices[utt_id] = matrix

            fc.close()

            ark_path = os.path.join(path, '{}.ark'.format(FEATS_FILE_NAME))
            ark_path = os.path.abspath(ark_path)
            scp_path = os.path.join(path, '{}.scp'.format(FEATS_FILE_NAME))

            self.write_float_matrices(scp_path, ark_path, matrices)

    @staticmethod
    def feature_scp_generator(path):
        """ Return a generator over all feature matrices defined in a scp. """

        scp_entries = textfile.read_key_value_lines(path, separator=' ')

        for utterance_id, rx_specifier in scp_entries.items():
            yield utterance_id, KaldiWriter.read_float_matrix(rx_specifier)

    @staticmethod
    def read_float_matrix(rx_specifier):
        """ Return float matrix as np array for the given rx specifier. """

        path, offset = rx_specifier.strip().split(':', maxsplit=1)
        offset = int(offset)
        sample_format = 4

        with open(path, 'rb') as f:
            # move to offset
            f.seek(offset)

            # assert binary ark
            binary = f.read(2)
            assert (binary == b'\x00B')

            # assert type float 32
            format = f.read(3)
            assert (format == b'FM ')

            # get number of mfcc features
            f.read(1)
            num_frames = struct.unpack('<i', f.read(4))[0]

            # get size of mfcc features
            f.read(1)
            feature_size = struct.unpack('<i', f.read(4))[0]

            # read feature data
            data = f.read(num_frames * feature_size * sample_format)

            feature_vector = np.frombuffer(data, dtype='float32')
            feature_matrix = np.reshape(
                feature_vector,
                (num_frames, feature_size)
            )

            return feature_matrix

    def _get_utt_idx(self, utt):
        if (self.prefix_utterances_with_speaker and
                utt.issuer is not None and
                not utt.idx.startswith(utt.issuer.idx)):
            return '{}-{}'.format(utt.issuer.idx, utt.idx)
        else:
            return utt.idx

    @staticmethod
    def write_float_matrices(scp_path, ark_path, matrices):
        """
        Write the given dict matrices (utt-id/float ndarray)
        to the given scp and ark files.
        """

        scp_entries = []

        with open(ark_path, 'wb') as f:
            for utterance_id in sorted(list(matrices.keys())):
                matrix = matrices[utterance_id]

                assert (matrix.dtype == np.float32)

                f.write(('{} '.format(utterance_id)).encode('utf-8'))

                offset = f.tell()

                f.write(b'\x00B')
                f.write(b'FM ')
                f.write(b'\x04')
                f.write(struct.pack('<i', np.size(matrix, 0)))
                f.write(b'\x04')
                f.write(struct.pack('<i', np.size(matrix, 1)))

                flattened = matrix.reshape(
                    np.size(matrix, 0) * np.size(matrix, 1)
                )
                flattened.tofile(f, sep='')

                scp_entries.append('{} {}:{}'.format(
                    utterance_id,
                    ark_path,
                    offset
                ))

        with open(scp_path, 'w') as f:
            f.write('\n'.join(scp_entries))



import os

from audiomate import corpus
from audiomate import issuers
from audiomate.corpus import io
from audiomate.corpus.io import voxforge

import pytest
import requests_mock

from tests import resources
from . import reader_test as rt


@pytest.fixture()
def sample_response():
    with open(resources.get_resource_path(['sample_files', 'voxforge_response.html']), 'r') as f:
        return f.read()


@pytest.fixture()
def sample_tgz_content():
    with open(resources.get_resource_path(['sample_files', 'voxforge_sample.tgz']), 'rb') as f:
        return f.read()


class TestVoxforgeDownloader:

    def test_download(self, tmpdir, sample_tgz_content):
        with requests_mock.Mocker() as mock:
            mock.get(voxforge.DOWNLOAD_URL['de'],
                     text='hallo <a href="test.tgz">test.tgz</a> blub')
            mock.get(os.path.join(voxforge.DOWNLOAD_URL['de'], 'test.tgz'), content=sample_tgz_content)

            downloader = voxforge.VoxforgeDownloader(lang='de')
            downloader.download(tmpdir.strpath)

        base_folder = os.path.join(tmpdir.strpath, 'Aaron-20080318-kdl')
        etc_folder = os.path.join(base_folder, 'etc')
        wav_folder = os.path.join(base_folder, 'wav')

        assert os.path.isfile(os.path.join(etc_folder, 'README'))

        assert os.path.isfile(os.path.join(wav_folder, 'b0019.wav'))
        assert os.path.isfile(os.path.join(wav_folder, 'b0020.wav'))
        assert os.path.isfile(os.path.join(wav_folder, 'b0021.wav'))
        assert os.path.isfile(os.path.join(wav_folder, 'b0022.wav'))
        assert os.path.isfile(os.path.join(wav_folder, 'b0023.wav'))
        assert os.path.isfile(os.path.join(wav_folder, 'b0024.wav'))
        assert os.path.isfile(os.path.join(wav_folder, 'b0025.wav'))
        assert os.path.isfile(os.path.join(wav_folder, 'b0026.wav'))
        assert os.path.isfile(os.path.join(wav_folder, 'b0027.wav'))
        assert os.path.isfile(os.path.join(wav_folder, 'b0028.wav'))

    def test_download_custom_url(self, tmpdir, sample_tgz_content):
        with requests_mock.Mocker() as mock:
            mock.get('http://someurl.com/some/download/dir',
                     text='hallo <a href="test.tgz">test.tgz</a> blub')
            mock.get('http://someurl.com/some/download/dir/test.tgz', content=sample_tgz_content)

            downloader = voxforge.VoxforgeDownloader(lang='de', url='http://someurl.com/some/download/dir')
            downloader.download(tmpdir.strpath)

        base_folder = os.path.join(tmpdir.strpath, 'Aaron-20080318-kdl')
        etc_folder = os.path.join(base_folder, 'etc')
        wav_folder = os.path.join(base_folder, 'wav')

        assert os.path.isfile(os.path.join(etc_folder, 'README'))

        assert os.path.isfile(os.path.join(wav_folder, 'b0019.wav'))
        assert os.path.isfile(os.path.join(wav_folder, 'b0020.wav'))
        assert os.path.isfile(os.path.join(wav_folder, 'b0021.wav'))
        assert os.path.isfile(os.path.join(wav_folder, 'b0022.wav'))
        assert os.path.isfile(os.path.join(wav_folder, 'b0023.wav'))
        assert os.path.isfile(os.path.join(wav_folder, 'b0024.wav'))
        assert os.path.isfile(os.path.join(wav_folder, 'b0025.wav'))
        assert os.path.isfile(os.path.join(wav_folder, 'b0026.wav'))
        assert os.path.isfile(os.path.join(wav_folder, 'b0027.wav'))
        assert os.path.isfile(os.path.join(wav_folder, 'b0028.wav'))

    def test_available_files(self, sample_response):
        with requests_mock.Mocker() as mock:
            url = 'http://someurl.com/some/download/dir'
            mock.get(url, text=sample_response)
            files = voxforge.VoxforgeDownloader.available_files(url)

            assert set(files) == {
                'http://someurl.com/some/download/dir/1337ad-20170321-amr.tgz',
                'http://someurl.com/some/download/dir/1337ad-20170321-bej.tgz',
                'http://someurl.com/some/download/dir/1337ad-20170321-blf.tgz',
                'http://someurl.com/some/download/dir/1337ad-20170321-czb.tgz',
                'http://someurl.com/some/download/dir/1337ad-20170321-hii.tgz'
            }

    def test_download_files(self, tmpdir):
        files = [
            'http://someurl.com/some/download/dir/1337ad-20170321-amr.tgz',
            'http://someurl.com/some/download/dir/1337ad-20170321-bej.tgz',
            'http://someurl.com/some/download/dir/1337ad-20170321-blf.tgz',
            'http://someurl.com/some/download/dir/1337ad-20170321-czb.tgz',
            'http://someurl.com/some/download/dir/1337ad-20170321-hii.tgz'
        ]

        with requests_mock.Mocker() as mock:
            for file in files:
                mock.get(file, content='some content'.encode())

            file_paths = voxforge.VoxforgeDownloader.download_files(files, tmpdir.strpath)

        expected_file_paths = [
            os.path.join(tmpdir.strpath, '1337ad-20170321-amr.tgz'),
            os.path.join(tmpdir.strpath, '1337ad-20170321-bej.tgz'),
            os.path.join(tmpdir.strpath, '1337ad-20170321-blf.tgz'),
            os.path.join(tmpdir.strpath, '1337ad-20170321-czb.tgz'),
            os.path.join(tmpdir.strpath, '1337ad-20170321-hii.tgz')
        ]

        assert len(file_paths) == 5

        for file_path in expected_file_paths:
            assert os.path.isfile(file_path)
            assert file_path in file_paths

    def test_extract_files(self, tmpdir):
        sample_file_path = resources.get_resource_path(['sample_files', 'voxforge_sample.tgz'])
        extracted = voxforge.VoxforgeDownloader.extract_files([sample_file_path], tmpdir.strpath)

        base_folder = os.path.join(tmpdir.strpath, 'Aaron-20080318-kdl')
        etc_folder = os.path.join(base_folder, 'etc')
        wav_folder = os.path.join(base_folder, 'wav')

        assert os.path.isfile(os.path.join(etc_folder, 'README'))

        assert os.path.isfile(os.path.join(wav_folder, 'b0019.wav'))
        assert os.path.isfile(os.path.join(wav_folder, 'b0020.wav'))
        assert os.path.isfile(os.path.join(wav_folder, 'b0021.wav'))
        assert os.path.isfile(os.path.join(wav_folder, 'b0022.wav'))
        assert os.path.isfile(os.path.join(wav_folder, 'b0023.wav'))
        assert os.path.isfile(os.path.join(wav_folder, 'b0024.wav'))
        assert os.path.isfile(os.path.join(wav_folder, 'b0025.wav'))
        assert os.path.isfile(os.path.join(wav_folder, 'b0026.wav'))
        assert os.path.isfile(os.path.join(wav_folder, 'b0027.wav'))
        assert os.path.isfile(os.path.join(wav_folder, 'b0028.wav'))

        assert len(extracted) == 1
        assert os.path.join(tmpdir.strpath, 'voxforge_sample') in extracted


class TestVoxforgeReader(rt.CorpusReaderTest):

    SAMPLE_PATH = resources.sample_corpus_path('voxforge')
    FILE_TRACK_BASE_PATH = SAMPLE_PATH

    EXPECTED_NUMBER_OF_TRACKS = 13
    EXPECTED_TRACKS = [
        rt.ExpFileTrack('1337ad-20170321-czb-de11-095', os.path.join('1337ad-20170321-czb',
                                                                     'wav',
                                                                     'de11-095.wav')),
        rt.ExpFileTrack('1337ad-20170321-czb-de11-096', os.path.join('1337ad-20170321-czb',
                                                                     'wav',
                                                                     'de11-096.wav')),
        rt.ExpFileTrack('anonymous-20081027-njq-a0479', os.path.join('anonymous-20081027-njq',
                                                                     'wav',
                                                                     'a0479.wav')),
    ]

    EXPECTED_NUMBER_OF_ISSUERS = 4
    EXPECTED_ISSUERS = [
        rt.ExpSpeaker('1337ad', 5, issuers.Gender.FEMALE, issuers.AgeGroup.ADULT, 'deu'),
        rt.ExpSpeaker('anonymous-20081027-njq', 1, issuers.Gender.MALE, issuers.AgeGroup.ADULT, 'eng'),
        rt.ExpSpeaker('Katzer', 4, issuers.Gender.MALE, issuers.AgeGroup.YOUTH, 'eng'),
        rt.ExpSpeaker('knotyouraveragejo', 3, issuers.Gender.FEMALE, issuers.AgeGroup.ADULT, 'eng'),
    ]

    EXPECTED_NUMBER_OF_UTTERANCES = 13
    EXPECTED_UTTERANCES = [
        rt.ExpUtterance('1337ad-20170321-czb-de11-095', '1337ad-20170321-czb-de11-095',
                        '1337ad', 0, float('inf')),
        rt.ExpUtterance('1337ad-20170321-czb-de11-096', '1337ad-20170321-czb-de11-096',
                        '1337ad', 0, float('inf')),
        rt.ExpUtterance('anonymous-20081027-njq-a0479', 'anonymous-20081027-njq-a0479',
                        'anonymous-20081027-njq', 0, float('inf')),
    ]

    EXPECTED_LABEL_LISTS = {
        '1337ad-20170321-czb-de11-096': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT_RAW, 1),
        ],
        'Katzer-20140410-lyk-b0167': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT_RAW, 1),
        ],
        'anonymous-20081027-njq-a0479': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT_RAW, 1),
        ],
    }

    EXPECTED_LABELS = {
        '1337ad-20170321-czb-de11-096': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT,
                        'ES HANDELT SICH UM GETRENNTE RECHTSSYSTEME UND NUR EINES IST ANWENDBAR',
                        0, float('inf')),
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT_RAW,
                        'Es handelt sich um getrennte Rechtssysteme und nur eines ist anwendbar.',
                        0, float('inf')),
        ],
        'Katzer-20140410-lyk-b0167': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT,
                        'A LITTLE BEFORE DAWN OF THE DAY FOLLOWING THE FIRE RELIEF CAME',
                        0, float('inf')),
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT_RAW,
                        'A little before dawn of the day following, the fire relief came.',
                        0, float('inf')),
        ],
        'anonymous-20081027-njq-a0479': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT,
                        'I TRIED TO READ GEORGE MOORE LAST NIGHT AND WAS DREADFULLY BORED',
                        0, float('inf')),
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT_RAW,
                        'I tried to read George Moore last night, and was dreadfully bored.',
                        0, float('inf')),
        ],
    }

    EXPECTED_NUMBER_OF_SUBVIEWS = 0

    def load(self):
        return io.VoxforgeReader().load(self.SAMPLE_PATH)
